

import numpy as np
import tensorflow as tf
import keras
from keras import Sequential
from keras import backend as K
from keras import regularizers
from keras.layers import (
    Activation,
    BatchNormalization,
    Conv2D,
    GlobalAveragePooling1D,
    Dense,
    Flatten,
    Lambda,
    Layer,
    Permute,
    maximum,
    Reshape,
)
from tensorflow.python.framework import tensor_shape
import keras
if keras.__version__.startswith("3"):
    import keras.ops as ops 
else:
    import tensorflow.keras.backend as ops



def ExpandDimension(dim: int = -1):
    return Lambda(lambda x: ops.expand_dims(x, dim))


def RemoveDimension(axis=0):
    return Lambda(lambda x: ops.squeeze(x, axis=axis))


class ConcreteDropout(Layer):
    """
    Concrete Dropout layer class from https://arxiv.org/abs/1705.07832.
    Dropout Feature Ranking for Deep Learning Models
    Chun-Hao Chang
    Ladislav Rampasek
    Anna Goldenberg

    Parameters:
        dropout_regularizer: Positive float, satisfying $dropout_regularizer = 2 / (\tau * N)$
            with model precision $\tau$ (inverse observation noise) and
            N the number of instances in the dataset.
            The factor of two should be ignored for cross-entropy loss,
            and used only for the eucledian loss.
        init_min: Minimum value for the randomly initialized dropout rate, in [0, 1].
        init_min: Maximum value for the randomly initialized dropout rate, in [0, 1],
            with init_min <= init_max.
        name: String, name of the layer.

    """

    def __init__(
        self,
        dropout_regularizer=1e-5,
        init_min=0.1,
        init_max=0.9,
        name=None,
        training=True,
        **kwargs,
    ):
        super(ConcreteDropout, self).__init__(name=name, **kwargs)
        assert init_min <= init_max, "init_min must be lower or equal to init_max."

        self.dropout_regularizer = dropout_regularizer

        self.p_logit = None
        self.p = None
        self.init_min = np.log(init_min) - np.log(1.0 - init_min)
        self.init_max = np.log(init_max) - np.log(1.0 - init_max)
        self.training = training

    def build(self, input_shape):
        self.window = input_shape[-2]
        self.number_of_features = input_shape[-1]
        input_shape = tensor_shape.TensorShape(input_shape)

        self.p_logit = self.add_weight(
            name="p_logit",
            shape=[self.number_of_features],
            initializer=tf.random_uniform_initializer(self.init_min, self.init_max),
            dtype=tf.float32,
            trainable=True,
        )

    def concrete_dropout(self, p, x):
        eps = K.cast_to_floatx(K.epsilon())
        temp = 1.0 / 10.0
        unif_noise = K.random_uniform(shape=[self.number_of_features])
        drop_prob = (
            K.log(p + eps)
            - K.log(1.0 - p + eps)
            + K.log(unif_noise + eps)
            - K.log(1.0 - unif_noise + eps)
        )
        drop_prob = K.sigmoid(drop_prob / temp)
        random_tensor = 1.0 - drop_prob

        retain_prob = 1.0 - p
        x *= random_tensor
        x /= retain_prob
        return x

    def call(self, inputs, training=True):
        p = K.sigmoid(self.p_logit)

        dropout_regularizer = p * K.log(p)
        dropout_regularizer += (1.0 - p) * K.log(1.0 - p)
        dropout_regularizer *= self.dropout_regularizer * self.number_of_features
        regularizer = K.sum(dropout_regularizer)
        self.add_loss(regularizer)

        x = self.concrete_dropout(p, inputs)

        return x


class ResidualShrinkageBlock(Layer):
    """
    ResidualShrinkageBlock

    """

    def build(self, input_shape):
        self.blocks = []
        for i in range(2):
            self.blocks.append(
                Sequential(
                    [
                        BatchNormalization(),
                        Activation("relu"),
                        Conv2D(1, (1, 1), padding="same"),
                    ]
                )
            )

        self.abs = Lambda(lambda x: tf.abs(x))
        self.abs_mean = Sequential(
            [
                RemoveDimension(3),
                Permute((1, 2)),
                GlobalAveragePooling1D(),
                ExpandDimension(2),
                ExpandDimension(3),
            ]
        )
        self.shrinkage = Sequential(
            [
                Flatten(),
                Dense(input_shape[-1]),
                BatchNormalization(),
                Activation("relu"),
                Dense(
                    input_shape[-1], kernel_regularizer=tf.keras.regularizers.l2(1e-4)
                ),
                Activation("sigmoid"),
                ExpandDimension(2),
                ExpandDimension(3),
            ]
        )

    def call(self, inputs):
        x = ExpandDimension()(inputs)

        x = self.blocks[0](x)

        residual = self.blocks[1](x)

        residual_abs = self.abs(residual)

        abs_mean = self.abs_mean(residual_abs)

        scales = self.shrinkage(abs_mean)
        thres = abs_mean * scales
        thres = Permute((2, 1, 3))(thres)
        sub = (residual_abs) - thres

        zeros = sub - sub

        n_sub = maximum([sub, zeros])

        residual = ops.sign(residual) * n_sub
        residual = RemoveDimension(3)(residual)
        return residual + inputs


class ZeroWeights(tf.keras.constraints.Constraint):
    def __init__(self, l1: float):
        self.l1 = l1

    def __call__(self, w):
        return tf.math.multiply(w, tf.cast(tf.abs(w) > self.l1, tf.float32))

    def get_config(self):
        return {"l1": self.l1}


class LASSOLayer(Layer):
    """
    LASSO Layer

    Parameters:
        l1: L1 regularization parameter
    """

    def __init__(self, l1: float):
        super(LASSOLayer, self).__init__()
        self.l1 = l1
        self.kernel_regularizer = regularizers.L1(l1)

    def build(self, input_shape):
        W_size = np.prod(input_shape[1:])
        self.w = self.add_weight(
            shape=(W_size,),
            initializer="random_normal",
            trainable=True,
            regularizer=self.kernel_regularizer,
            constraint=ZeroWeights(self.l1),
        )

        self.input_reshape = Reshape((W_size,))
        self.output_reshape = Reshape(input_shape[1:])

    def call(self, inputs):
        x = self.input_reshape(inputs)

        x = tf.math.multiply(self.w, x)

        self.add_metric(
            tf.math.reduce_sum(tf.cast(tf.abs(self.w) > 0, tf.float32)),
            name="Number of features",
        )
        return self.output_reshape(x)





import numpy as np
import pandas as pd
import pytest
from ceruleo.transformation.features.imputers import (ForwardFillImputer,
                                                       MeanImputer, ApplyRollingImputer,
                                                       MedianImputer, NaNtoInf)


class TestImputers():
    def test_ApplyRollingImputer(self):
        def finite_max(x):
            return np.max(x[np.isfinite(x)])
        f = ApplyRollingImputer(window_size=3, func=finite_max)
        df = pd.DataFrame({
            'a': [0, np.inf, 12, -np.inf,  0.9, 15, 0.5, 0.3, 0.5],
            'b': [5, 6,   7,   5,   np.inf,   5,  6,   5,   45],
        })
        assert not pd.isnull(df['a'][1])
        assert not pd.isnull(df['a'][2])
        assert not pd.isnull(df['b'][4])
        f.partial_fit(df)
        df_new = f.transform(df)

        assert np.all(np.isfinite(df_new))
        assert np.all(df_new['a'] == np.array([0, 12, 12, 15, 0.9, 15, 0.5, 0.3, 0.5]))

    def test_PandasRemoveInf(self):

        remover = NaNtoInf()
        df = pd.DataFrame({
            'a': [0, np.inf, -np.inf, 0.1, 0.9, 15, 0.5, 0.3, 0.5],
            'b': [5, 6,   7,   5,   np.inf,   5,  6,   5,   45],
        })
        assert not pd.isnull(df['a'][1])
        assert not pd.isnull(df['a'][2])
        assert not pd.isnull(df['b'][4])

        df_new = remover.fit_transform(df)

        assert pd.isnull(df_new['a'][1])
        assert pd.isnull(df_new['a'][2])
        assert pd.isnull(df_new['b'][4])

    def test_PandasMedianImputer(self):

        remover = MedianImputer()
        df = pd.DataFrame({
            'a': [0, np.nan, np.nan, 0.1, 0.9, 15, 0.5, 0.3, 0.5],
            'b': [5, 6,   np.nan,   5,   9,   5,  6,   5,   45],
        })
        a_median = np.median([0, 0.1, 0.9, 15, 0.5, 0.3, 0.5])
        b_median = np.median([5, 6,   5,   9,   5,  6,   5,   45])

        df_new = remover.fit_transform(df)

        assert not (pd.isnull(df_new).any().any())
        assert df_new['a'][1] ==  pytest.approx(a_median)
        assert df_new['a'][2] == pytest.approx(a_median)
        assert df_new['b'][2] == pytest.approx(b_median)

    def test_PandasMeanImputer(self):

        remover = MeanImputer()
        df = pd.DataFrame({
            'a': [0, np.nan, np.nan, 0.1, 0.9, 15, 0.5, 0.3, 0.5],
            'b': [5, 6,   np.nan,   5,   9,   5,  6,   5,   45],
        })
        a_mean = np.mean([0, 0.1, 0.9, 15, 0.5, 0.3, 0.5])
        b_mean = np.mean([5, 6,   5,   9,   5,  6,   5,   45])

        df_new = remover.fit_transform(df)
        assert not (pd.isnull(df_new).any().any())

        assert df_new['a'][1] == pytest.approx(a_mean)
        assert df_new['a'][2] == pytest.approx(a_mean)

        assert df_new['b'][2] == pytest.approx(b_mean)

    def test_ForwardFillImputer(self):

        remover = ForwardFillImputer()
        df = pd.DataFrame({
            'a': [0, 0.5, np.nan, np.nan, 0.9, 15, 0.5, 0.3, 0.5],
            'b': [5, 6,   7,   5,   np.nan,   5,  6,   5,   45],
        })
        df_new = remover.fit_transform(df)
        assert not (pd.isnull(df_new).any().any())

        assert df_new['a'][1] == pytest.approx(0.5)
        assert df_new['a'][2] == pytest.approx(0.5)

        assert df_new['b'][4] == pytest.approx(5)



"""Transformer step is the base class of all transformers

The pipeline will use the steps to fit and transform the run-to-failure
cycles
"""
from copy import copy
from typing import List, Optional

import pandas as pd
from ceruleo.transformation.functional.mixin import TransformerStepMixin
from sklearn.base import TransformerMixin


class TransformerStep(TransformerStepMixin, TransformerMixin):
    """Base class of all transformation step

    """
    def partial_fit(self, X:pd.DataFrame, y=None) -> "TransformerStep":
        """Fit a single run-to-failure cycle

        Parameters:

            X: Features of the run-to-failure cycle


        Returns:
            TransformerStep: The same step
        """
        return self

    def fit(self, X, y=None)  -> "TransformerStep":
        """Fit the complete set of run-to-failure cycles

        Parameters:

            X: Features of the all the run-to-failure cycles


        Returns:
            TransformerStep: The same step
        """
        return self

    def find_feature(self, X: pd.DataFrame, name: str) -> Optional[str]:
        """Find the feature that best maches the columns in X

        Parameters:
            X: A run-to-failure cycle
            name: The name of the feature to find

        Returns:
            The name of the columns if it was found, else None
        
        """
        matches = [c for c in X.columns if name in c]
        if len(matches) > 0:
            return matches[0]
        else:
            return None

    def description(self):
        return f"{self.name}"




    def __add__(self, other):
        from ceruleo.transformation.features.operations import Sum
        from ceruleo.transformation.utils import ensure_step
        return Sum()([self, ensure_step(other)])

    def __truediv__(self, other):
        from ceruleo.transformation.features.operations import Divide
        from ceruleo.transformation.utils import ensure_step
        return Divide()([self, ensure_step(other)])



from collections import OrderedDict

import numpy as np
from ceruleo.models.torch.model import TorchTrainableModel

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import optim


class Single_Global_SelfAttn_Module(nn.Module):

    def __init__(
            self,
            window, n_multiv, n_kernels, w_kernel,
            d_k, d_v, d_model, d_inner,
            n_layers, n_head, drop_prob=0.1):
        '''
        Args:
        window (int): the length of the input window size
        n_multiv (int): num of univariate time series
        n_kernels (int): the num of channels
        w_kernel (int): the default is 1
        d_k (int): d_model / n_head
        d_v (int): d_model / n_head
        d_model (int): outputs of dimension
        d_inner (int): the inner-layer dimension of Position-wise Feed-Forward Networks
        n_layers (int): num of layers in Encoder
        n_head (int): num of Multi-head
        drop_prob (float): the probability of dropout
        '''

        super(Single_Global_SelfAttn_Module, self).__init__()

        self.window = window
        self.w_kernel = w_kernel
        self.n_multiv = n_multiv
        self.d_model = d_model
        self.drop_prob = drop_prob
        self.conv2 = nn.Conv2d(1, n_kernels, (window, w_kernel))
        self.in_linear = nn.Linear(n_kernels, d_model)
        self.out_linear = nn.Linear(d_model, n_kernels)

        self.layer_stack = nn.ModuleList([
            EncoderLayer(d_model, d_inner, n_head, d_k, d_v, dropout=drop_prob)
            for _ in range(n_layers)])

    def forward(self, x, return_attns=False):

        x = x.view(-1, self.w_kernel, self.window, self.n_multiv)
        x2 = F.relu(self.conv2(x))
        x2 = nn.Dropout(p=self.drop_prob)(x2)
        x = torch.squeeze(x2, 2)
        x = torch.transpose(x, 1, 2)
        src_seq = self.in_linear(x)

        enc_slf_attn_list = []

        enc_output = src_seq

        for enc_layer in self.layer_stack:
            enc_output, enc_slf_attn = enc_layer(enc_output)
            if return_attns:
                enc_slf_attn_list += [enc_slf_attn]

        if return_attns:
            return enc_output, enc_slf_attn_list
        enc_output = self.out_linear(enc_output)
        return enc_output,


class Single_Local_SelfAttn_Module(nn.Module):

    def __init__(
            self,
            window, local, n_multiv, n_kernels, w_kernel,
            d_k, d_v, d_model, d_inner,
            n_layers, n_head, drop_prob=0.1):
        '''
        Args:
        window (int): the length of the input window size
        n_multiv (int): num of univariate time series
        n_kernels (int): the num of channels
        w_kernel (int): the default is 1
        d_k (int): d_model / n_head
        d_v (int): d_model / n_head
        d_model (int): outputs of dimension
        d_inner (int): the inner-layer dimension of Position-wise Feed-Forward Networks
        n_layers (int): num of layers in Encoder
        n_head (int): num of Multi-head
        drop_prob (float): the probability of dropout
        '''

        super(Single_Local_SelfAttn_Module, self).__init__()

        self.window = window
        self.w_kernel = w_kernel
        self.n_multiv = n_multiv
        self.d_model = d_model
        self.drop_prob = drop_prob
        self.conv1 = nn.Conv2d(1, n_kernels, (local, w_kernel))
        self.pooling1 = nn.AdaptiveMaxPool2d((1, n_multiv))
        self.in_linear = nn.Linear(n_kernels, d_model)
        self.out_linear = nn.Linear(d_model, n_kernels)

        self.layer_stack = nn.ModuleList([
            EncoderLayer(d_model, d_inner, n_head, d_k, d_v, dropout=drop_prob)
            for _ in range(n_layers)])

    def forward(self, x, return_attns=False):

        x = x.view(-1, self.w_kernel, self.window, self.n_multiv)
        x1 = F.relu(self.conv1(x))
        x1 = self.pooling1(x1)
        x1 = nn.Dropout(p=self.drop_prob)(x1)
        x = torch.squeeze(x1, 2)
        x = torch.transpose(x, 1, 2)
        src_seq = self.in_linear(x)

        enc_slf_attn_list = []

        enc_output = src_seq

        for enc_layer in self.layer_stack:
            enc_output, enc_slf_attn = enc_layer(enc_output)
            if return_attns:
                enc_slf_attn_list += [enc_slf_attn]

        if return_attns:
            return enc_output, enc_slf_attn_list
        enc_output = self.out_linear(enc_output)
        return enc_output,


class AR(nn.Module):

    def __init__(self, window: int, hidden_size: int):

        super(AR, self).__init__()
        self.linear = nn.Linear(window, hidden_size)

    def forward(self, x):
        x = torch.transpose(x, 1, 2)
        x = self.linear(x)
        x = torch.transpose(x, 1, 2)
        return x


class ScaledDotProductAttention(nn.Module):
    ''' Scaled Dot-Product Attention '''

    def __init__(self, temperature, attn_dropout=0.1):
        super().__init__()
        self.temperature = temperature
        self.dropout = nn.Dropout(attn_dropout)
        self.softmax = nn.Softmax(dim=2)

    def forward(self, q, k, v):

        attn = torch.bmm(q, k.transpose(1, 2))
        attn = attn / self.temperature

        attn = self.softmax(attn)
        attn = self.dropout(attn)
        output = torch.bmm(attn, v)

        return output, attn


class MultiHeadAttention(nn.Module):
    ''' Multi-Head Attention module '''

    def __init__(self, n_head, d_model, d_k, d_v, dropout=0.1):
        super().__init__()

        self.n_head = n_head
        self.d_k = d_k
        self.d_v = d_v

        self.w_qs = nn.Linear(d_model, n_head * d_k)
        self.w_ks = nn.Linear(d_model, n_head * d_k)
        self.w_vs = nn.Linear(d_model, n_head * d_v)
        nn.init.normal_(self.w_qs.weight, mean=0,
                        std=np.sqrt(2.0 / (d_model + d_k)))
        nn.init.normal_(self.w_ks.weight, mean=0,
                        std=np.sqrt(2.0 / (d_model + d_k)))
        nn.init.normal_(self.w_vs.weight, mean=0,
                        std=np.sqrt(2.0 / (d_model + d_v)))

        self.attention = ScaledDotProductAttention(
            temperature=np.power(d_k, 0.5))
        self.layer_norm = nn.LayerNorm(d_model)

        self.fc = nn.Linear(n_head * d_v, d_model)
        nn.init.xavier_normal_(self.fc.weight)

        self.dropout = nn.Dropout(dropout)

    def forward(self, q, k, v):

        d_k, d_v, n_head = self.d_k, self.d_v, self.n_head

        sz_b, len_q, _ = q.size()
        sz_b, len_k, _ = k.size()
        sz_b, len_v, _ = v.size()

        residual = q

        q = self.w_qs(q).view(sz_b, len_q, n_head, d_k)
        k = self.w_ks(k).view(sz_b, len_k, n_head, d_k)
        v = self.w_vs(v).view(sz_b, len_v, n_head, d_v)

        q = q.permute(2, 0, 1, 3).contiguous().view(-1, len_q, d_k)
        k = k.permute(2, 0, 1, 3).contiguous().view(-1, len_k, d_k)
        v = v.permute(2, 0, 1, 3).contiguous().view(-1, len_v, d_v)

        output, attn = self.attention(q, k, v)

        output = output.view(n_head, sz_b, len_q, d_v)
        output = output.permute(1, 2, 0, 3).contiguous().view(sz_b, len_q, -1)

        output = self.dropout(self.fc(output))
        output = self.layer_norm(output + residual)

        return output, attn


class PositionwiseFeedForward(nn.Module):
    ''' A two-feed-forward-layer module '''

    def __init__(self, d_in, d_hid, dropout=0.1):
        super().__init__()
        self.w_1 = nn.Conv1d(d_in, d_hid, 1)
        self.w_2 = nn.Conv1d(d_hid, d_in, 1)
        self.layer_norm = nn.LayerNorm(d_in)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        residual = x
        output = x.transpose(1, 2)
        output = self.w_2(F.relu(self.w_1(output)))
        output = output.transpose(1, 2)
        output = self.dropout(output)
        output = self.layer_norm(output + residual)
        return output


class EncoderLayer(nn.Module):
    ''' Compose with two layers '''

    def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1):
        super(EncoderLayer, self).__init__()
        self.slf_attn = MultiHeadAttention(
            n_head, d_model, d_k, d_v, dropout=dropout)
        self.pos_ffn = PositionwiseFeedForward(
            d_model, d_inner, dropout=dropout)

    def forward(self, enc_input):
        enc_output, enc_slf_attn = self.slf_attn(
            enc_input, enc_input, enc_input)

        enc_output = self.pos_ffn(enc_output)

        return enc_output, enc_slf_attn


class DecoderLayer(nn.Module):
    ''' Compose with three layers '''

    def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1):
        super(DecoderLayer, self).__init__()
        self.slf_attn = MultiHeadAttention(
            n_head, d_model, d_k, d_v, dropout=dropout)
        self.enc_attn = MultiHeadAttention(
            n_head, d_model, d_k, d_v, dropout=dropout)
        self.pos_ffn = PositionwiseFeedForward(
            d_model, d_inner, dropout=dropout)

    def forward(self, dec_input, enc_output, non_pad_mask=None, slf_attn_mask=None, dec_enc_attn_mask=None):
        dec_output, dec_slf_attn = self.slf_attn(
            dec_input, dec_input, dec_input, mask=slf_attn_mask)

        dec_output, dec_enc_attn = self.enc_attn(
            dec_output, enc_output, enc_output, mask=dec_enc_attn_mask)

        dec_output = self.pos_ffn(dec_output)

        return dec_output, dec_slf_attn, dec_enc_attn


class DSANetPrivate(nn.Module):

    def __init__(self, model: TorchTrainableModel, local: int = 2, n_kernels: int = 64,
                 w_kernel: int = 1, d_model: int = 5, d_inner: int = 6, n_layers: int = 4,
                 n_head: int = 8, d_k: int = 3, d_v: int = 3, dropout: float = 0.3, hidden_size: int = 16):
        """
        Pass in parsed HyperOptArgumentParser to the model
        """
        super(DSANetPrivate, self).__init__()

        self.local = local
        self.n_kernels = n_kernels
        self.w_kernel = w_kernel

        # hyperparameters of model
        self.d_model = d_model
        self.d_inner = d_inner
        self.n_layers = n_layers
        self.n_head = n_head
        self.d_k = d_k
        self.d_v = d_v
        self.drop_prob = dropout
        self.hidden_size = hidden_size

        self.model = model

        self.build_model()

    def build_model(self):
        """
        Layout model
        """
        self.sgsf = Single_Global_SelfAttn_Module(
            window=self.model.window, n_multiv=self.model.n_features, n_kernels=self.n_kernels,
            w_kernel=self.w_kernel, d_k=self.d_k, d_v=self.d_v, d_model=self.d_model,
            d_inner=self.d_inner, n_layers=self.n_layers, n_head=self.n_head, drop_prob=self.drop_prob)

        self.slsf = Single_Local_SelfAttn_Module(
            window=self.model.window, local=self.local, n_multiv=self.model.n_features, n_kernels=self.n_kernels,
            w_kernel=self.w_kernel, d_k=self.d_k, d_v=self.d_v, d_model=self.d_model,
            d_inner=self.d_inner, n_layers=self.n_layers, n_head=self.n_head, drop_prob=self.drop_prob)

        self.ar = AR(window=self.model.window, hidden_size=self.hidden_size)
        self.W_output1 = nn.Linear(2 * self.n_kernels, self.hidden_size)
        self.dropout = nn.Dropout(p=self.drop_prob)
        self.active_func = nn.Tanh()

        self.final_layer = nn.Linear(self.hidden_size*self.model.n_features, 1)

    def forward(self, x):

        sgsf_output, *_ = self.sgsf(x)
        slsf_output, *_ = self.slsf(x)

        sf_output = torch.cat((sgsf_output, slsf_output), 2)
        sf_output = self.dropout(sf_output)
        sf_output = self.W_output1(sf_output)

        sf_output = torch.transpose(sf_output, 1, 2)

        ar_output = self.ar(x)

        output = sf_output + ar_output

        output = output.reshape(-1, self.model.n_features*self.hidden_size)

        output = F.relu(self.final_layer(output))
        output = output.view(-1, 1, 1)
        return output

    def training_step(self, data_batch, batch_i):
        """
        Lightning calls this inside the training loop
        """
        x, y = data_batch

        y_hat = self.forward(x)

        # calculate loss
        loss_val = self.loss(y, y_hat)

        # in DP mode (default) make sure if result is scalar, there's another dim in the beginning
        if self.trainer.use_dp:
            loss_val = loss_val.unsqueeze(0)

        output = OrderedDict({
            'loss': loss_val
        })

        # can also return just a scalar instead of a dict (return loss_val)
        return output

    def validation_step(self, data_batch, batch_i):
        """
        Lightning calls this inside the validation loop
        """
        x, y = data_batch

        y_hat = self.forward(x)

        loss_val = self.loss(y, y_hat)

        # in DP mode (default) make sure if result is scalar, there's another dim in the beginning
        if self.trainer.use_dp:
            loss_val = loss_val.unsqueeze(0)

        output = OrderedDict({
            'val_loss': loss_val,
            'y': y,
            'y_hat': y_hat,
        })

        # can also return just a scalar instead of a dict (return loss_val)
        return output

    def validation_end(self, outputs):
        """
        Called at the end of validation to aggregate outputs
        :param outputs: list of individual outputs of each validation step
        """
        # if returned a scalar from validation_step, outputs is a list of tensor scalars
        # we return just the average in this case (if we want)
        # return torch.stack(outputs).mean()

        loss_sum = 0
        for x in outputs:
            loss_sum += x['val_loss'].item()
        val_loss_mean = loss_sum / len(outputs)

        y = torch.cat(([x['y'] for x in outputs]), 0)
        y_hat = torch.cat(([x['y_hat'] for x in outputs]), 0)

        num_var = y.size(-1)
        y = y.view(-1, num_var)
        y_hat = y_hat.view(-1, num_var)
        sample_num = y.size(0)

        y_diff = y_hat - y
        y_mean = torch.mean(y)
        y_translation = y - y_mean

        val_rrse = torch.sqrt(torch.sum(torch.pow(y_diff, 2))) / \
            torch.sqrt(torch.sum(torch.pow(y_translation, 2)))

        y_m = torch.mean(y, 0, True)
        y_hat_m = torch.mean(y_hat, 0, True)
        y_d = y - y_m
        y_hat_d = y_hat - y_hat_m
        corr_top = torch.sum(y_d * y_hat_d, 0)
        corr_bottom = torch.sqrt(
            (torch.sum(torch.pow(y_d, 2), 0) * torch.sum(torch.pow(y_hat_d, 2), 0)))
        corr_inter = corr_top / corr_bottom
        val_corr = (1. / num_var) * torch.sum(corr_inter)

        val_mae = (1. / (sample_num * num_var)) * torch.sum(torch.abs(y_diff))

        tqdm_dic = {
            'val_loss': val_loss_mean,
            'RRSE': val_rrse.item(),
            'CORR': val_corr.item(),
            'MAE': val_mae.item()
        }
        return tqdm_dic


class DSANet(TorchTrainableModel):
    def __init__(self, local: int = 7, n_kernels: int = 64, w_kernel: int = 1,
                 d_model: int = 512, d_inner: int = 2048, n_layers: int = 6, n_head: int = 8,
                 d_k: int = 64, d_v: int = 64, dropout: float = 0.3, **kwargs):
        super(DSANet, self).__init__(**kwargs)
        self.model_params = [
            local, n_kernels, w_kernel, d_model, d_inner,
            n_layers, n_head, d_k, d_v, dropout

        ]

    def build_model(self):
        return DSANetPrivate(self, *self.model_params)

    def build_optimizer(self):
        return optim.Adam(
            self.model.parameters(), lr=self.learning_rate)

    def build_scheduler(self):
        return optim.lr_scheduler.CosineAnnealingLR(self.optimizer, T_max=10)



from typing import List, Optional, Tuple
from ceruleo.models.keras.layers import ExpandDimension
from tensorflow.keras import Input, Model, optimizers
from tensorflow.keras.layers import Conv2D, Dense, Dropout, Flatten
import tensorflow as tf


def XiangQiangJianQiaoModel(
    input_shape: Tuple[int, int],
    *,
    convolutional_layers: List[Tuple[int, int]] = [(10, 10), (10, 10), (10, 10)],
    dropout: float = 0.5,
    dense_dimension: int = 100,
    convolutional_activation: str = "tanh",
) -> tf.keras.Model:
    """Model presented in "Remaining useful life estimation in prognostics using deep convolution neural networks"


    Xiang Li, Qian Ding, Jian-Qiao Sun


    Deafult parameters reported in the article:

        Number of filters:	10
        Window size:	30/20/30/15
        Filter length: 10

        Neurons in fully-connected layer:	100
        Dropout rate:	0.5
        batch_size: 512

    Parameters:

        input_shape: input shape of the iterator
            
        convolutional_layers : List[Tuple[int, int]], optional
            List of tuples with the convolutional dimensions
            Each element of the list is a tuple that contains
            (number of filter, kernel size)
        dropout: Dropout rate
        dense_dimension: Dimension of the fully connected layer
        convolutional_activation: Activation of the convolutional layers

    Return:
    
        model: tf.keras.Model
        
    """

    input = Input(shape=input_shape)
    x = input

    x = ExpandDimension()(x)
    for n_filters, filter_size in convolutional_layers:
        x = Conv2D(
            n_filters,
            (filter_size, 1),
            padding="same",
            activation=convolutional_activation,
        )(x)

    x = Conv2D(1, (3, 1), padding="same", activation=convolutional_activation)(x)

    x = Flatten()(x)
    x = Dropout(dropout)(x)
    x = Dense(dense_dimension, activation=convolutional_activation)(x)
    output = Dense(1, activation="relu")(x)
    model = Model(
        inputs=[input],
        outputs=[output],
    )
    return model



from typing import Tuple

from tensorflow.keras.layers import (
    Concatenate,
    Conv2D,
    Dense,
    Dropout,
    Flatten,
    Permute,
    Reshape,
)
from tensorflow.keras import Input, Model


def MVCNN(
    input_shape: Tuple[int, int],
    *,
    window = 64,
    dropout: float=0.1,
    
):
    """
    Model presented in Remaining useful life estimation in prognostics using deep convolution neural networks

    Li, X., Ding, Q., & Sun, J. Q. (2018).
    Remaining useful life estimation in prognostics using deep convolution neural networks.
    Reliability Engineering & System Safety, 172, 1-11.

    Deafult parameters reported in the article:

        Number of filters:	10
        Window size:	30/20/30/15
        Filter length: 10

        Neurons in fully-connected layer	100
        Dropout rate	0.5
        batch_size = 512


    Parameters:

        window: Window size of the convolutional windows
        dropout: 
        
    """
    n_features = input_shape[1]
    window = input_shape[0]

    input = Input(shape=input_shape)
    x = input
    x = Permute((2, 1))(x)
    x = Reshape((input_shape[0], input_shape[1], window))(x)

    x = Conv2D(window, (1, 1), activation="relu", padding="same")(x)

    x1 = Conv2D(window, (2, 2), activation="relu", padding="same")(x)
    x1 = Conv2D(window, (2, 2), activation="relu", padding="same")(x1)

    x2 = Conv2D(window, (3, 3), activation="relu", padding="same")(x)
    x2 = Conv2D(window, (3, 3), activation="relu", padding="same")(x2)

    x3 = Conv2D(window, (5, 5), activation="relu", padding="same")(x)
    x3 = Conv2D(window, (5, 5), activation="relu", padding="same")(x3)

    x = Concatenate(axis=1)([x, x1, x2, x3])
    x = Conv2D(window, input_shape)(x)
    x = Flatten()(x)
    x = Dense(100, activation="relu")(x)
    x = Dropout(dropout)(x)
    x = Dense(100, activation="relu")(x)
    x = Dropout(dropout)(x)
    output = Dense(1)(x)
    model = Model(
        inputs=[input],
        outputs=[output],
    )
    return model



from collections.abc import Iterable
from pathlib import Path



try: 
    from types import EllipsisType
except:
    EllipsisType = type(Ellipsis)
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

try:
    import tensorflow as tf

    TENSORFLOW_ENABLED = True
except:
    TENSORFLOW_ENABLED = False
from abc import ABC, abstractmethod, abstractproperty

from tqdm.auto import tqdm


class DatasetIterator:
    def __init__(self, dataset):
        self.dataset = dataset
        self.i = 0

    def __next__(self):
        if self.i == self.dataset.n_time_series:
            raise StopIteration
        a = self.dataset[self.i]
        self.i += 1
        return a


class AbstractPDMDataset(ABC):
    def __init__(self):
        self._common_features = None
        self._durations = None

    def __iter__(self):
        return DatasetIterator(self)

    @abstractproperty
    @property
    def n_time_series(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_time_series(self, i: int) -> pd.DataFrame:
        """

        Returns:
            df: DataFrame with the data of the life i
        """
        raise NotImplementedError

    def number_of_samples_of_time_series(self, i: int) -> int:
        return self[i].shape[0]

    @abstractproperty
    def rul_column(self) -> str:
        raise NotImplementedError

    def duration(self, life: pd.DataFrame) -> float:
        """Obtain the duration of the time-series

        Parameters:
            life: The input life

        Returns:
            Duration of the life
        """
        return life[self.rul_column].max()

    def number_of_samples(self) -> List[int]:
        return [
            self.number_of_samples_of_time_series(i) for i in tqdm(range(len(self)))
        ]



    def durations(self, show_progress: bool = False) -> List[float]:
        """
        Obtain the length of each life

        Return:
            List of durations
        """
        if self._durations is None:
            if show_progress:
                iterator = tqdm(self)
            else:
                iterator = self
            self._durations = [self.duration(life) for life in iterator]
            # [self.rul_column].iloc[0]
        return self._durations

    def __call__(self, i):
        return self[i]

    def get_features_of_life(self, i: int) -> pd.DataFrame:
        return self[i]


    def __getitem__(
        self, i: Union[int, Iterable]
    ) -> Union[pd.DataFrame, "FoldedDataset"]:
        """Obtain a time-series or an splice of the dataset using a FoldedDataset

        Parameters:
            i: If the parameter is an int it will return a pd.DataFrame with the i-th time-series.
                If the parameter is a list of int it will return a FoldedDataset with the
                time-series whose id are present in the list

        Raises:
            ValueError: When the list does not contain integer parameters

        Returns:
            The i-th time-series
            An instance of class FoldedDataset containing the dataset with the lives specified by the list
        """
        if isinstance(i, slice):
            i = range(
                0 if i.start is None else i.start,
                len(self) if i.stop is None else i.stop,
                1 if i.step is None else i.step,
            )
        if TENSORFLOW_ENABLED and isinstance(i, tf.Tensor):
            return self.get_time_series(i.ref())
        


        if isinstance(i, Iterable):
            if not all(isinstance(item, (int, np.integer)) for item in i):
                if len(i) == 2:       
                    if not isinstance(i[1], EllipsisType):
                        raise ValueError("Invalid iterable index passed")
                    i = i[0]        

            return FoldedDataset(self, i)
        else:
            df = self.get_time_series(i)
            return df

    @property
    def shape(self) -> Tuple[int]:
        return (self.n_time_series,)

    def __len__(self) -> int:
        """
        Compute the number of lifes in the dataset

        Return:
            Number of time-series in the dataset
        """
        return self.n_time_series

    def to_pandas(
        self,
        proportion_of_lives: float = 1.0,
        subsample_proportion: float = 1.0,
        show_progress: bool = False,
    ) -> pd.DataFrame:
        """
        Create a dataset with the time-series concatenated

        Parameters:
            proportion_of_lives: Proportion of lives to use, by default 1
            subsample_proportion: Proportion of samples to use, by default 1
            show_progress: Whether to show progress when concatenating the lives, by default False

        Returns:
            A DataFrame with all the lives concatenated
        """
        if show_progress:
            bar = tqdm
        else:
            bar = lambda x: x
        df = []

        features = list(
            self._compute_common_features(
                proportion_of_lives=proportion_of_lives, show_progress=show_progress
            )
        )

        for i in bar(range(self.n_time_series)):
            if proportion_of_lives < 1.0 and np.random.rand() > proportion_of_lives:
                continue

            current_life = self[i].loc[:, features]
            if subsample_proportion < 1.0:
                indices = range(
                    0,
                    current_life.shape[0],
                    int(current_life.shape[0] * subsample_proportion),
                )
                current_life = current_life.iloc[indices, :]
            df.append(current_life)
        return pd.concat(df)

    def _compute_common_features(
        self, proportion_of_lives: float = 1.0, show_progress: bool = False
    ) -> List[str]:
        common_features = []
        if show_progress:
            bar = tqdm
        else:
            bar = lambda x: x
        for i in bar(range(self.n_time_series)):
            if proportion_of_lives < 1.0 and np.random.rand() > proportion_of_lives:
                continue
            life = self.get_features_of_life(i)
            common_features.append(set(life.columns.values))
        return sorted(list(common_features[0].intersection(*common_features)))

    def common_features(
        self, show_progress: bool = False, proportion_of_lives: float = 1.0
    ) -> List[str]:
        """
        Compute the common features of the dataset among the different lives

        Parameters:
            proportion_of_lives: Proportion of lives to use, by default 1
            show_progress: Whether to show progress when computing the common features, by default False

        Returns:
            A list with the common features
        """
        if self._common_features is None:
            self._common_features = self._compute_common_features(
                proportion_of_lives, show_progress=show_progress
            )
        return self._common_features

    def map(
        self, transformer: "TransformedDataset", cache_size: int = None
    ) -> "TransformedDataset":
        """
        Apply a transformation to the dataset

        Parameters:
            transformer: The transformation to apply
            cache_size: The size of the cache to use, by default None

        Returns:
            The transformed dataset as an instance of class TransformedDataset
        """
        from ceruleo.dataset.transformed import TransformedDataset

        return TransformedDataset(self, transformer, cache_size=cache_size)

    def numeric_features(self, show_progress: bool = False) -> List[str]:
        """Obtain the list of the common numeric features in the dataset

        Parameters:
            show_progress: Whether to show progress when computing the common features, by default False

        Returns:
            List of columns containing the common numeric features
        """

        features = self.common_features(show_progress=show_progress)
        df = self.get_features_of_life(0)
        return list(
            df.loc[:, features]
            .select_dtypes(include=[np.number], exclude=["datetime", "timedelta"])
            .columns.values
        )

    def categorical_features(self, show_progress: bool = False) -> List[str]:
        """Obtain the list of the common categorical features in the dataset

        Parameters:
            show_progress: Whether to show progress when computing the common features

        Returns:
            List of columns containing the common numeric features
        """
        features = self.common_features(show_progress=show_progress)
        df = self.get_time_series(0)
        return list(
            df.loc[:, features]
            .select_dtypes(exclude=[np.number, "datetime", "timedelta"])
            .columns.values
        )


class FoldedDataset(AbstractPDMDataset):
    """
    Dataset containing a subset of the time-series. An instanc of this class can be obtained by slicing an AbstractTimeSeriesDataset with a list of indexes
    """

    def __init__(self, dataset: AbstractPDMDataset, indices: list):
        super().__init__()
        self.dataset = dataset
        self.indices = indices

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except Exception as e:
            return self.dataset.__getattribute__(__name)

    @property
    def n_time_series(self) -> int:
        """
        Compute the number of lifes in the folded dataset

        Return:
            Number of time-series in the folded dataset

        """
        return len(self.indices)

    def get_time_series(self, i: int) -> pd.DataFrame:
        """
        Obtain the i-th time-series in the folded dataset

        Parameters:
            i: Index of the life

        Returns:
            The i-th time-series
        """
        return self.dataset[self.indices[i]]

    def _original_index(self, i: int) -> int:
        """
        Obtain the index of the i-th time-series in the original dataset

        Parameters:
            i: Index of the life

        Returns:
            The index of the i-th time-series in the original dataset
        """
        if isinstance(self.dataset, FoldedDataset):
            return self.dataset._original_index(self.indices[i])
        else:
            return self.indices[i]

    def original_indices(self) -> List[int]:
        """
        Obtain the original indices for all the time-series in the FoldedDataset

        Returns:
            The original indices for all the time-series in the FoldedDataset
        """
        return [self._original_index(i) for i in range(len(self.indices))]

    def number_of_samples_of_time_series(self, i: int) -> int:
        """
        Compute the number of samples of the i-th time-series in the FoldedDataset

        Parameters:
            i: Index of the life

        Returns:
            Number of samples of the i-th time-series in the FoldedDataset
        """
        return self[i][0].shape[0]

    def __reduce_ex__(self, __protocol) -> Union[str, Tuple[Any, ...]]:
        return (self.__class__, (self.dataset, self.indices))

    @property
    def rul_column(self):
        return self.dataset.rul_column


class PDMDataset(AbstractPDMDataset):
    dataset_path: Path 
    procesed_path: Path 
    cycles_table_filename: Path 
    cycles_metadata: pd.DataFrame
    _rul_column: str


    def __init__(self, path: Path, rul_column: str = "RUL"):
        super().__init__()
        self.dataset_path = path        
        self.procesed_path = self.dataset_path / "processed" / "cycles"
        self.cycles_table_filename = self.procesed_path / "cycles.csv"
        self._prepare_dataset()
        self.cycles_metadata = pd.read_csv(self.cycles_table_filename)
        self._rul_column = rul_column

    def _prepare_dataset(self):
        pass

    def get_time_series(self, i: int) -> pd.DataFrame:
        file_path = Path(self.cycles_metadata.iloc[i]["Filename"]).resolve()
        file_extension = file_path.suffix.lower()
        if file_extension == '.csv':
            return pd.read_csv(file_path)
        elif file_extension in ('.xls', '.xlsx'):
            return pd.read_excel(file_path)
        elif file_extension == '.parquet':
            return pd.read_parquet(file_path)
        elif file_extension == '.feather':
            return pd.read_feather(file_path)
        elif file_extension in ('.h5', '.hdf5'):
            return pd.read_hdf(file_path)
        elif file_extension in ('.pkl', '.pickle'):
            return pd.read_pickle(file_path)
        else:
            raise ValueError(f"Unsupported file extension: {file_extension}")


    
    @property
    def n_time_series(self) -> int:
        return len(self.cycles_metadata)

    @property
    def rul_column(self) -> int:
        return self._rul_column


class PDMInMemoryDataset(AbstractPDMDataset):
    cycles: List[pd.DataFrame]
    _rul_column: str
    cycles_metadata: Optional[pd.DataFrame]

    def __init__(
        self,
        cycles: List[pd.DataFrame],
        rul_column: str,
        cycles_metadata: Optional[pd.DataFrame] = None,
    ):
        super().__init__()
        self.cycles = cycles
        self._rul_column = rul_column
        self.cycles_metadata = cycles_metadata
        

    def get_time_series(self, i: int) -> pd.DataFrame:
        return self.cycles[i]

    @property
    def n_time_series(self) -> int:
        return len(self.cycles)

    @property
    def rul_column(self) -> int:
        return self._rul_column



from concurrent.futures import ProcessPoolExecutor
from typing import Callable, List, Optional, Union

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.tdigest import TDigest
from sklearn.pipeline import FeatureUnion, _transform_one


class PandasToNumpy(TransformerStep):
    def transform(self, X):
        return X.values


class TransformerLambda(TransformerStep):
    def __init__(
        self, *, callback: Callable[[pd.DataFrame], pd.DataFrame], name: Optional[str] = None
    ):
        super().__init__(name=name)
        self.f = callback

    def transform(self, X, y=None):
        return self.f(X)


class IdentityTransformerStep(TransformerStep):
    def transform(self, input_array, y=None):
        if isinstance(input_array, pd.DataFrame):
            return input_array.copy()
        else:
            return input_array * 1


class SKLearnTransformerWrapper(TransformerStep):
    def __init__(self, transformer, name: Optional[str] = None):
        super().__init__(name)
        self.transformer = transformer

    def fit(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        self.transformer.fit(X.values)
        return self

    def partial_fit(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        if hasattr(self.transformer, "partial_fit"):
            self.transformer.partial_fit(X.values)
        else:
            self.transformer.fit(X.values)
        return self

    def _column_names(self, X) -> List[str]:
        return X.columns

    def transform(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        return pd.DataFrame(
            self.transformer.transform(X), columns=self._column_names(X), index=X.index
        )


def column_names_window(columns: list, window: int) -> list:
    """

    Parameters
    ----------
    columns: list
             List of column names

    window: int
            Window size

    Return
    ------
    Column names with the format: w_{step}_{feature_name}
    """
    new_columns = []
    for w in range(1, window + 1):
        for c in columns:
            new_columns.append(f"w_{w}_{c}")
    return new_columns


def build_tdigest(tdigest, values, column):
    return column, tdigest.merge_sorted(values)


class QuantileEstimator:
    """Approximate the quantile of each feature in the dataframe
    using t-digest
    """

    def __init__(
        self,
        tdigest_size: int = 200,
        max_workers: int = 1,
        subsample: Optional[Union[int, float]] = None,
    ):
        self.tdigest_dict = None
        self.tdigest_size = tdigest_size
        self.max_workers = max_workers
        self.subsample = subsample

    def update(self, X: pd.DataFrame):
        if X.shape[0] < 2:
            return self

        columns = X.columns

        if self.tdigest_dict is None:
            self.tdigest_dict = {c: TDigest(self.tdigest_size) for c in columns}

        results = []
        with ProcessPoolExecutor(max_workers=self.max_workers) as executor:
            for i, c in enumerate(columns):

                x = X.iloc[:, i].dropna()
                if self.subsample is not None:

                    if isinstance(self.subsample, int):
                        points_to_sample = self.subsample
                    else:
                        points_to_sample = self.subsample * X.shape[0]
                    step = max(int(X.shape[0] / float(points_to_sample)), 1)
                    x = x.iloc[::step]
                x = x.sort_values().values
                t_digest = self.tdigest_dict[c]
                results.append(executor.submit(build_tdigest, t_digest, x, c))

        for r in results:
            c, tdigest = r.result()
            self.tdigest_dict[c] = tdigest
        return self

    def estimate_quantile(self, *args, **kwargs):
        return self.quantile(*args, **kwargs)

    def quantile(
        self, q: float, feature: Optional[str] = None
    ) -> Union[pd.Series, float]:
        """Estimate the quantile for a set of features

        Parameters
        ----------
        q:float
          The quantile to estimate
        feature:Optional[Str]"""
        if feature is not None:
            return self.tdigest_dict[feature].estimate_quantile(q)
        else:
            return pd.Series(
                {
                    c: self.tdigest_dict[c].estimate_quantile(q)
                    for c in self.tdigest_dict.keys()
                }
            )


class Literal(TransformerStep):
    def __init__(self, literal, *args):
        super().__init__(*args)
        self.literal = literal

    def transform(self, X):
        return self.literal


def ensure_step(step):
    if isinstance(step, TransformerStep):
        return step
    return Literal(step)



import requests
from tqdm.auto import tqdm
from pathlib import Path 

def download(URL:str, output_path: Path):
    response = requests.get(URL, stream=True)
    total_size_in_bytes= int(response.headers.get('content-length', 0))
    block_size = 1024 
    progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
    with open(output_path, 'wb') as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()


from typing import List, Optional, Tuple, Type


import matplotlib.pyplot as plt
import matplotlib.figure
import numpy as np
from matplotlib.colors import LogNorm, Normalize
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np


def XCM_explanation(
    features_explanations: np.ndarray, times_explanations: np.ndarray, *, cmap="bwr"
) -> matplotlib.figure.Figure:
    """Plot the explanations of the XCM model

    Parameters

        features_explanations: Features explanations provided by from ceruleo.models.keras.catalog.XCM.explain
        times_explaantion: Times explanations provided by from ceruleo.models.keras.catalog.XCM.explain
        cmap: Colormap

    Return

        fig: matplotlib Figure
    """

    fig, ax = plt.subplots(1, 2, figsize=(17, 5))
    ax[0].set_title("Feature importance")
    im1 = ax[0].imshow(features_explanations.T, cmap=cmap)
    ax[0].set_ylabel("Features")
    ax[0].set_xlabel("Time")
    ax[0].grid(None)
    divider = make_axes_locatable(ax[0])
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(im1, cax=cax, orientation="vertical")

    ax[1].set_title("Temporal importance")
    im2 = ax[1].imshow(
        np.repeat(np.expand_dims(times_explanations, 1), 3, axis=1).T, cmap=cmap
    )
    ax[1].set_yticks([])
    ax[1].grid(None)
    ax[1].set_xlabel("Time")
    divider = make_axes_locatable(ax[1])
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(im2, cax=cax, orientation="vertical")
    return fig


def show_LASSOLayer(
    model, iterator_shape: Tuple[int, int], *, scale: bool = True, **fig_kwargs
):
    """Shows the LASSO layer

    Parameters

        model: tf.keras.Model with a LassoLayer at the beggining
        iterator_shape: Input shape of the model

    Returns

        fig: matplotlib.figure.Figure
    """
    im = np.abs(model.layers[1].w.numpy().reshape(iterator_shape).T)
    if scale:
        im = (im - im.min()) / (im.max() - im.min())

    fig, ax = plt.subplots(**fig_kwargs)
    c = ax.imshow(im, interpolation="nearest", aspect="auto")
    fig.colorbar(c)
    return fig



import functools
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from ceruleo.graphics.duration import durations_boxplot, durations_histogram
import pytest
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.graphics.analysis import plot_correlation_analysis
from ceruleo.graphics.explanations import XCM_explanation
from ceruleo.graphics.results import (
    barplot_errors_wrt_RUL,
    boxplot_errors_wrt_RUL,
    plot_predictions,
    plot_predictions_grid,
    shadedline_plot_errors_wrt_RUL,
)
from ceruleo.results.results import PredictionResult

np.random.seed(19680801)

PATH = Path(__file__).resolve().parent
TEST_IMAGES_PATH = PATH / "test_images"


def helper_test_plot(output_filename: str):
    def inner_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            print(args, kwargs)
            (TEST_IMAGES_PATH).mkdir(parents=True, exist_ok=True)
            figure = func(*args, **kwargs)
            EXPECTED_IMAGE_PATH = TEST_IMAGES_PATH / (
                output_filename + "-expected" + ".png"
            )
            if not (EXPECTED_IMAGE_PATH).is_file():
                figure.savefig(EXPECTED_IMAGE_PATH)
                assert False
            CURRENT_IMAGE_PATH = TEST_IMAGES_PATH / (output_filename + ".png")
            figure.savefig(CURRENT_IMAGE_PATH)
            assert (
                np.mean(
                    np.abs(
                        plt.imread(EXPECTED_IMAGE_PATH) - plt.imread(CURRENT_IMAGE_PATH)
                    )
                )
                < 0.1
            )

        return wrapper

    return inner_decorator


def create_predictions(name: str, number_of_lives: int) -> PredictionResult:
    y_trues = []
    y_preds = []
    for i in range(number_of_lives):
        N = np.random.randint(1500) + 200
        y_true = np.linspace(500, 0, N)
        s = np.random.rand() * 500
        y_pred = y_true + np.random.randn(N) * s
        y_trues.append(y_true)
        y_preds.append(y_pred)

    y_true = np.hstack((*y_trues,))
    y_pred = np.hstack((*y_preds,))
    return PredictionResult(name, y_true, y_pred)


class MockDataset(AbstractPDMDataset):
    def __init__(self, nlives: int):
        super().__init__()
        self.lives = [
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, 100, 50),
                    "feature2": np.random.randint(2, size=(50,)),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
            for i in range(nlives)
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


@pytest.fixture(scope="class")
def predictions():
    return {
        "Model A": [create_predictions("Model A", 5), create_predictions("Model A", 5)],
        "Model B": [create_predictions("Model B", 5), create_predictions("Model B", 8)],
    }


class TestGraphics:
    @helper_test_plot(output_filename="test_plot_predictions_grid_1")
    def test_plot_predictions_grid_1(self):
        y_true = np.linspace(500, 0, num=500)
        y_pred = y_true + np.random.rand(500) * 15
        r = PredictionResult("Example", y_true, y_pred)
        ax = plot_predictions_grid(r, ncols=1)
        return ax[0][0].figure

    @helper_test_plot(output_filename="test_plot_predictions_grid_2")
    def test_plot_predictions_grid_2(self):
        y_true = np.linspace(500, 0, num=500)
        y_pred = y_true + np.random.rand(500) * 15
        y_true = np.hstack((y_true, y_true))
        y_pred = np.hstack((y_pred, y_pred))
        r = PredictionResult("Example", y_true, y_pred)
        ax = plot_predictions_grid(r, ncols=2)
        return ax[0][0].figure

    @helper_test_plot(output_filename="test_plot_predictions_grid_2")
    def test_plot_predictions_grid_3(self):
        y_true = np.linspace(500, 0, num=500)
        y_pred = y_true + np.random.rand(500) * 15
        y_true = np.hstack((y_true, y_true))
        y_pred = np.hstack((y_pred, y_pred))
        r = PredictionResult("Example", y_true, y_pred)
        ax = plot_predictions_grid(r, ncols=1)
        return ax[0][0].figure

    @helper_test_plot(output_filename="test_boxplot_errors_wrt_RUL")
    def test_boxplot_errors_wrt_RUL(self, predictions):
        ax = boxplot_errors_wrt_RUL(predictions, nbins=5)
        return ax.figure

    @helper_test_plot(
        output_filename="test_barplot_errors_wrt_RUL",
    )
    def test_barplot_errors_wrt_RUL(self, predictions):
        ax = barplot_errors_wrt_RUL(predictions, nbins=5)
        return ax.figure

    @helper_test_plot(
        output_filename="test_shadedline_plot_errors_wrt_RUL",
    )
    def test_shadedline_plot_errors_wrt_RUL(self, predictions):
        ax = shadedline_plot_errors_wrt_RUL(predictions, nbins=5)
        return ax.figure

    @helper_test_plot(
        output_filename="test_XCM_explanation",
    )
    def test_XCM_explanation(self):
        mmap = np.random.rand(40, 40)
        mtime = np.random.rand(40)
        fig = XCM_explanation(mmap, mtime)
        return fig

    @helper_test_plot(
        output_filename="test_correlation_plot",
    )
    def test_correlation_plot(self):
        dataset = MockDataset(7)
        ax = plot_correlation_analysis(dataset)
        return ax.figure

    @helper_test_plot(
        output_filename="test_durations_boxplot",
    )
    def test_durations_boxplot(self):
        dataset = MockDataset(7)
        ax = durations_boxplot(dataset, xlabel="Test Dataset")
        return ax.figure

    @helper_test_plot(
        output_filename="test_durations_histogram",
    )
    def test_durations_histogram(self):
        dataset = MockDataset(7)
        ax = durations_histogram(dataset, label="Test Dataset")
        return ax.figure



import functools
import gzip
import pickle
from multiprocessing import Pool
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.transformation.functional.transformers import Transformer
from ceruleo.utils.lrucache import LRUDataCache
from sklearn.utils.validation import check_is_fitted
from tqdm.auto import tqdm


def _transform(transformer, dataset, i: int):
    data = dataset[i]
    return (i, transformer.transform(data))


class TransformedDataset(AbstractPDMDataset):
    def __init__(
        self, dataset, transformer: Transformer, cache_size: Optional[int] = None
    ):
        super().__init__()
        self.transformer = transformer
        self.dataset = dataset
        if cache_size is None:
            cache_size = len(dataset)
        self.cache = LRUDataCache(cache_size)
        check_is_fitted(transformer)

    @property
    def n_time_series(self) -> int:
        return self.dataset.n_time_series

    def __call__(self, i: int):
        return self[i]
    
    @property
    def rul_column(self):
        return self.dataset.rul_column

    def number_of_samples_of_time_series(self, i: int) -> int:
        _, y, _ = self[i]
        return y.shape[0]

    def preload(self):
        transform = functools.partial(_transform, self.transformer, self.dataset)
        with Pool(6) as p:
            values = list(
                tqdm(
                    p.imap(transform, range(self.n_time_series)),
                    total=self.n_time_series,
                    desc="Preloading",
                )
            )
        for i, (X, y, metadata) in values:
            self.cache.add(i, (X, y, metadata))

    def get_time_series(self, i: int) -> pd.DataFrame:
        if i not in self.cache.data:
            data = self.dataset[i]
            X, y, metadata = self.transformer.transform(data)
            self.cache.add(i, (X, y, metadata))
        X, y, metadata = self.cache.get(i)
        return X, y, metadata

    def get_features_of_life(
        self, i: int, pandas: bool = True
    ) -> Union[np.ndarray, pd.DataFrame]:
        X, _, _ = self[i]
        if pandas:
            return X
        else:
            return X.values

    def save(self, output_path: Path):
        TransformedSerializedDataset.save(self, output_path)


class TransformedSerializedDataset(TransformedDataset):
    @staticmethod
    def save(dataset: TransformedDataset, output_path: Path):
        if not output_path.is_dir():
            output_path.mkdir(parents=True, exist_ok=True)
        file_list = {}
        for i, life in enumerate(dataset):
            file_list[i] = f"ts_{i}.pkl.gz"
            with gzip.open(output_path / f"ts_{i}.pkl.gz", "wb") as file:
                pickle.dump(life, file)
        with open(output_path / "transformer.pkl", "wb") as file:
            pickle.dump(dataset.transformer, file)
        with open(output_path / "lives.pkl", "wb") as file:
            pickle.dump(file_list, file)

    def __init__(self, dataset_path: Path, cache_size: Optional[int] = None):
        self.dataset_path = dataset_path
        with open(dataset_path / "lives.pkl", "rb") as file:
            self.files = pickle.load(file)
        with open(dataset_path / "transformer.pkl", "rb") as file:
            self.transformer = pickle.load(file)
        if cache_size is None:
            cache_size = len(self.files)
        self.cache = LRUDataCache(cache_size)

    def _open_file(self, i: int):
        with gzip.open(self.dataset_path / self.files[i], "rb") as file:
            return pickle.load(file)

    def get_time_series(self, i: int) -> pd.DataFrame:
        if i not in self.cache.data:
            X, y, metadata = self._open_file(i)
            self.cache.add(i, (X, y, metadata))
        return self.cache.get(i)

    @property
    def n_time_series(self):
        return len(self.files)

    def __len__(self):
        return self.n_time_series




import logging
from multiprocessing import JoinableQueue, Manager, Process, Queue
from typing import Iterable, Union

import pandas as pd
from ceruleo.transformation.functional.graph_utils import (
    root_nodes,
    topological_sort_iterator,
)
from ceruleo.transformation.functional.pipeline.cache_store import CacheStoreType
from ceruleo.transformation.functional.pipeline.traversal import CachedGraphTraversal
from ceruleo.transformation.functional.transformerstep import TransformerStep
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)


def _transform(node, old_element, dataset_element, queue):
    n = node.transform(old_element)
    queue.put((dataset_element, n))


class CachedPipelineRunner:
    """Performs an execution of the transformation graph caching the intermediate results

    Parameters:

        final_step: Last step of the graph
        cache_type: Mode for storing the cache
    """

    def __init__(
        self,
        final_step: TransformerStep,
        cache_type: CacheStoreType = CacheStoreType.SHELVE,
    ):

        self.final_step = final_step
        self.root_nodes = root_nodes(final_step)
        self.cache_type = cache_type

    def _run(
        self,
        dataset: Iterable[pd.DataFrame],
        fit: bool = True,
        show_progress: bool = False,
    ):
        dataset_size = len(dataset)

        with CachedGraphTraversal(
            self.root_nodes, dataset, cache_type=self.cache_type
        ) as cache:
            for node in topological_sort_iterator(self.final_step):
                if isinstance(node, TransformerStep) and fit:
                    if node.prefer_partial_fit:
                        for dataset_element in range(dataset_size):
                            d = cache.state_up_to(node, dataset_element)
                            node.partial_fit(d)
                    else:
                        data = pd.concat(
                            [
                                cache.state_up_to(node, dataset_element)
                                for dataset_element in range(dataset_size)
                            ]
                        )
                        node.fit(data)

                dataset_size = len(dataset)
                # if dataset_size > 1:
                # self._parallel_transform_step(cache, node, dataset_size, show_progress)
                # else:
                self._transform_step(cache, node, dataset_size, show_progress)

            last_state_key = cache.get_keys_of(None)[0]
            return cache.transformed_cache[last_state_key]

    def fit(self, dataset: Iterable[pd.DataFrame], show_progress: bool = False):
        return self._run(dataset, fit=True, show_progress=show_progress)

    def _update_step(self, cache, node, dataset_element, new_element):
        cache.clean_state_up_to(node, dataset_element)

        if len(node.next) > 0:
            for n in node.next:
                cache.store(n, node, dataset_element, new_element)
        else:
            cache.store(None, node, dataset_element, new_element)

    def _parallel_transform_step(
        self, cache: CachedGraphTraversal, node, dataset_size: int, show_progress: bool
    ):
        if show_progress:
            bar = tqdm(range(dataset_size))
            bar.set_description(node.name)
        else:
            bar = range(dataset_size)

        producers = []

        queue = JoinableQueue(dataset_size)
        for dataset_element in range(dataset_size):
            old_element = cache.state_up_to(
                node,
                dataset_element,
            )
            producers.append(
                Process(
                    target=_transform, args=(node, old_element, dataset_element, queue)
                )
            )
        for p in producers:
            p.start()

        for _ in bar:
            dataset_element, new_element = queue.get()
            queue.task_done()
            self._update_step(cache, node, dataset_element, new_element)
        cache.remove_state(node)
        for p in producers:
            p.join()
        queue.join()

    def _transform_step(
        self, cache: CachedGraphTraversal, node, dataset_size: int, show_progress: bool
    ):
        if show_progress:
            bar = tqdm(range(dataset_size))
            bar.set_description(node.name)
        else:
            bar = range(dataset_size)
        try:
            for dataset_element in bar:
                old_element = cache.state_up_to(node, dataset_element)
                new_element = node.transform(old_element)
                self._update_step(cache, node, dataset_element, new_element)
        except Exception as e:
            logger.error(f"There was an error when transforming with {node.name}")
            raise

    def transform(self, df: Union[pd.DataFrame, Iterable[pd.DataFrame]]):
        if isinstance(df, pd.DataFrame):
            return self._run([df], fit=False)
        else:
            return self._run(df, fit=False)



import tensorflow as tf

from ceruleo.iterators.iterators import WindowedDatasetIterator
import numpy as np


def tf_regression_dataset(iterator: WindowedDatasetIterator) -> tf.data.Dataset:
    """
    Create a forecast tf.data.Dataset from the iterator

    The dataset is constructed from a generator

    Parameters:
        iterator: The data iterator

    Returns:
        A tensorflow dataset
    """
    n_features = iterator.n_features

    def generator_function():
        for X, y, sw in iterator:
            yield X, y, sw

    a = tf.data.Dataset.from_generator(
        generator_function,
        output_signature=(
            tf.TensorSpec(shape=(iterator.window_size, n_features), dtype=tf.float32),
            tf.TensorSpec(shape=(iterator.horizon, 1), dtype=tf.float32),
            tf.TensorSpec(shape=(1), dtype=tf.float32),
        ),
    )

    return a


def tf_seq_to_seq_dataset(iterator: WindowedDatasetIterator) -> tf.data.Dataset:
    """
    Create a sequence to sequence tf.data.Dataset from the iterator

    The dataset is constructed from a generator

    Parameters:
        iterator: The data iterator

    Returns:
        A tensorflow dataset
    """
    n_features = iterator.n_features

    def generator_function():
        for X, y, sw in iterator:
            yield X, y, sw

    a = tf.data.Dataset.from_generator(
        generator_function,
        output_signature=(
            tf.TensorSpec(shape=(iterator.window_size, n_features), dtype=tf.float32),
            tf.TensorSpec(
                shape=(iterator.window_size, iterator.horizon), dtype=tf.float32
            ),
            tf.TensorSpec(shape=(1), dtype=tf.float32),
        ),
    )

    return a


def tf_autoencoder_dataset(iterator: WindowedDatasetIterator) -> tf.data.Dataset:
    """
    Create an autoencoder tf.data.Dataset from the iterator

    The dataset is constructed from a generator

    Parameters:
        iterator: The data iterator

    Returns:
        A tensorflow dataset
    """
    n_features = iterator.n_features

    def gen_train():
        for X, y, sw in iterator:
            yield X, X, sw

    a = tf.data.Dataset.from_generator(
        gen_train,
        (tf.float32, tf.float32, tf.float32),
        (
            tf.TensorShape([iterator.window_size, n_features]),
            tf.TensorShape([iterator.window_size, n_features]),
            tf.TensorShape([1]),
        ),
    )

    return a



import numpy as np
import tensorflow as tf
from scipy.special import loggamma
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Concatenate, Dense, Lambda, Multiply


class TFWeibullDistribution:
    @staticmethod
    def log_likelihood(x: tf.Tensor, alpha: tf.Tensor, beta: tf.Tensor):
        ya = (x + 0.00000000001) / alpha
        return tf.reduce_mean(K.log(beta) + (beta * K.log(ya)) - K.pow(ya, beta))

    @staticmethod
    def kl_divergence(l1: tf.Tensor, k1: tf.Tensor, l2: tf.Tensor, k2: tf.Tensor):
        term_1 = K.log(k1 / K.pow(l1, k1))
        term_2 = K.log(k2 / K.pow(l2, k2))
        term_3 = (k1 - k2) * (K.log(l1) - 0.5772 / k1)
        term_4 = K.pow((l1 / l2), k2) * K.exp(tf.math.lgamma((k2 / k1) + 1))
        tf.print(term_1, term_2, term_3, term_4)
        return K.mean(term_1 - term_2 + term_3 + term_4 - 1)


class WeibullDistribution:
    @staticmethod
    def mean(alpha, beta):
        return alpha * np.exp(loggamma(1 + (1 / beta)))

    @staticmethod
    def mode(alpha, beta):
        vmode = alpha * np.power((beta - 1) / beta, 1 / beta)
        vmode[beta <= 1] = 0
        return vmode

    @staticmethod
    def median(alpha, beta):
        return alpha * (np.power(np.log(2.0), (1 / beta)))

    @staticmethod
    def variance(alpha, beta):
        return alpha ** 2 * (
            np.exp(loggamma(1 + (2 / beta))) - (np.exp(loggamma(1 + (1 / beta)))) ** 2
        )

    @staticmethod
    def quantile(q, alpha, beta):
        return alpha * np.power(-np.log(1 - q), 1 / beta)


class NotCensoredWeibull(tf.keras.losses.Loss):
    def __init__(self, regression_weight: float = 5, likelihood_weight:float = 1):
        super().__init__()
        self.regression_weight = regression_weight
        self.likelihood_weight = likelihood_weight

    def call(self, y_true, y_pred):
        pRUL = y_pred[:, 0]
        alpha = y_pred[:, 1]
        beta = y_pred[:, 2]
        y_true = tf.squeeze(y_true)

        reg_loss = tf.keras.losses.MeanAbsoluteError()(pRUL, y_true)

        log_liks = TFWeibullDistribution.log_likelihood(y_true, alpha, beta)

        # log_liks = K.clip(log_liks, K.log(0.0000000001), K.log(1 - 0.0000000001))
        # + kl_weibull(alpha, beta, alpha, 2.0 )
        loss = -self.likelihood_weight*log_liks + self.regression_weight * reg_loss
        # + K.pow(ya,beta)
        return loss


class WeibullLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        return_params=True,
        regression="mode",
        name="WeibullParams",
        *args,
        **kwargs
    ):
        super().__init__(name=name, *args, **kwargs)
        self.return_params = return_params
        if self.return_params:
            self.params = Concatenate(name="Weibullparams")
        if regression == "mode":
            self.fun = self.mode
        elif regression == "mean":
            self.fun = self.mean
        elif regression == "median":
            self.fun = self.median

    def mean(self, lambda_pipe, k_pipe):
        inner_gamma = Lambda(lambda x: tf.math.exp(tf.math.lgamma(1 + (1 / x))))(k_pipe)
        return Multiply(name="RUL")([lambda_pipe, inner_gamma])

    def median(self, lambda_pipe, k_pipe):
        return lambda_pipe * (tf.math.pow(tf.math.log(2.0), tf.math.reciprocal(k_pipe)))

    def mode(self, alpha, beta):
        mask = K.cast(K.greater(beta, 1), tf.float32)
        beta = tf.clip_by_value(beta, 1 + 0.00000000001, np.inf)
        return mask * alpha * tf.math.pow((beta - 1) / beta, (1 / beta))

    def _result(self, alpha, beta):
        RUL = self.fun(alpha, beta)
        if self.return_params:
            return self.params([alpha, beta])
        else:
            return RUL


class WeibullParameters(WeibullLayer):
    def __init__(self, hidden, regression="mode", return_params=True, *args, **kwargs):
        super(WeibullParameters, self).__init__(
            return_params=True, regression=regression, name="", *args, **kwargs
        )

        self.W = Dense(hidden, activation="relu")

        self.xalpha1 = Dense(hidden, activation="relu")
        self.xalpha2 = Dense(1, name="w_alpha", activation='softplus')

        self.xbeta1 = Dense(hidden, activation="relu")
        self.xbeta2 = Dense(1, name="w_beta", activation='softplus')

    def call(self, input_tensor, training=False):

        x = self.W(input_tensor)

        alpha = self.xalpha1(x)
        alpha = self.xalpha2(alpha)

        beta = self.xbeta1(x)
        beta = self.xbeta2(beta)

        RUL = self.mode(alpha, beta)

        x = Concatenate(axis=1)([RUL, alpha, beta])

        return x



import math

import numpy as np
from ceruleo.iterators.iterators import RelativeToEnd, RelativeToStart, WindowedDatasetIterator
from ceruleo.iterators.shufflers import (
    AbstractShuffler,
    AllShuffled,
    IntraTimeSeriesShuffler,
    InverseOrder,
    NotShuffled,
    TimeSeriesOrderIntraSignalShuffling,
    TimeSeriesOrderShuffling,
)


class MockDataFrame:
    def __init__(self, size):
        self.size = size

    @property
    def shape(self):
        return [self.size, None]


class MockDataset:
    def __init__(self):
        self.sizes = [5, 7, 3]

    def __getitem__(self, id: int):
        return MockDataFrame(self.sizes[id])

    def number_of_samples_of_time_series(self, id: int):
        return self.sizes[id]

    @property
    def n_time_series(self):
        return 3


class MockDatasetBig:
    def __init__(self, N:int = 6046):
        self.sizes = [55, 32, 59, 125, N]

    def __getitem__(self, id: int):
        return MockDataFrame(self.sizes[id])

    def number_of_samples_of_time_series(self, id: int):
        return self.sizes[id]

    @property
    def n_time_series(self):
        return 5


class MockDatasetMedium:
    def __init__(self):
        self.sizes = [155, 132, 159, 125, 604]

    def __getitem__(self, id: int):
        return MockDataFrame(self.sizes[id])

    def number_of_samples_of_time_series(self, id: int):
        return self.sizes[id]

    @property
    def n_time_series(self):
        return 5     


class MockIterator:
    def __init__(self,  step: int = 1, dataset=MockDataset(), last_point:bool = True):
        self.dataset = dataset
        self.step = step
        self.start_index = RelativeToStart(0)
        self.end_index = RelativeToEnd(0)
        self.last_point = last_point


class TestShufflers:
    def test_shufflers(self):

        x = NotShuffled()
        generated = [e for e in x.iterator(MockIterator())]
        expected = [
            (0, 0),
            (0, 1),
            (0, 2),
            (0, 3),
            (0, 4),
            (1, 0),
            (1, 1),
            (1, 2),
            (1, 3),
            (1, 4),
            (1, 5),
            (1, 6),
            (2, 0),
            (2, 1),
            (2, 2),
        ]
        for g, e in zip(generated, expected):
            assert g == e

        x = NotShuffled()
        generated = [e for e in x.iterator(MockIterator(2))]
        expected = [
            (0, 0),
            (0, 2),
            (0, 4),
            (1, 0),
            (1, 2),
            (1, 4),
            (1, 6),
            (2, 0),
            (2, 2),
        ]
        for g, e in zip(generated, expected):
            assert g == e

    def test_intra_singal(self):
        np.random.seed(0)
        x = IntraTimeSeriesShuffler()
        iterator = MockIterator()
        generated = [e for e in x.iterator(iterator)]
        for i in range(3):
            samples_ts = [e for e in generated if e[0] == i]
            assert len(samples_ts) == iterator.dataset.sizes[i]
            ts_timestamps = [e[1] for e in samples_ts]
            assert ts_timestamps != list(range(0, iterator.dataset.sizes[i]))
            assert sorted(ts_timestamps) == list(range(0, iterator.dataset.sizes[i]))

        x = IntraTimeSeriesShuffler()
        step = 2
        iterator = MockIterator(step)
        generated = [e for e in x.iterator(iterator)]
        for i in range(3):
            samples_ts = [e for e in generated if e[0] == i]
            assert len(samples_ts) == math.ceil(iterator.dataset.sizes[i] / step)
            ts_timestamps = [e[1] for e in samples_ts]
            assert sorted(ts_timestamps) == list(
                np.arange(0, iterator.dataset.sizes[i], step=step)
            )

    def test_time_series_order(self):
        x = TimeSeriesOrderShuffling()
        it = MockIterator()
        generated = [e for e in x.iterator(it)]
        g = [g[0] for g in generated]
        assert len(np.where(np.diff(g))[0]) == 2
        i = [elem[1] for elem in generated if elem[0] == 2]
        assert i == list(range(0, it.dataset.sizes[2]))

        step = 2
        x = TimeSeriesOrderShuffling()
        it = MockIterator(step)
        generated = [e for e in x.iterator(it)]
        g = [g[0] for g in generated]
        assert len(np.where(np.diff(g))[0]) == 2
        i = [elem[1] for elem in generated if elem[0] == 2]
        assert i == np.arange(0, it.dataset.sizes[2], step=step).tolist()

    def test_TimeSeriesOrderIntraSignalShuffling(self):
        x = TimeSeriesOrderIntraSignalShuffling()
        it = MockIterator()
        generated = [e for e in x.iterator(it)]
        g = [g[0] for g in generated]
        assert len(np.where(np.diff(g))[0]) == 2

        step = 2
        x = TimeSeriesOrderIntraSignalShuffling()
        it = MockIterator(step)
        generated = [e for e in x.iterator(it)]
        g = [g[0] for g in generated]
        assert len(np.where(np.diff(g))[0]) == 2

    def test_InverseOrder(self):
        x = InverseOrder()
        it = MockIterator()
        generated = [e for e in x.iterator(it)]
        g = [g[1] for g in generated]
        assert np.all(np.diff(g) <= 0)

        step = 2
        x = InverseOrder()
        it = MockIterator(step)
        generated = [e for e in x.iterator(it)]
        g = [g[1] for g in generated]
        assert np.all(np.diff(g) <= 0)
        assert generated == [(1, 6), (0, 4), (1, 4), (0, 2), (1, 2), (2, 2), (0, 0), (1, 0), (2, 0)]

    def test_AllShuffled(self):
        x = AllShuffled()
        it = MockIterator()
        generated = [e for e in x.iterator(it)]
        assert len(generated) == np.sum(it.dataset.sizes)
        for i in range(3):
            g = [elem[0] for elem in generated if elem[0] == i]
            assert len(g) == it.dataset.sizes[i]

        step = 2
        x = AllShuffled()
        it = MockIterator(step)
        generated = [e for e in x.iterator(it)]
        assert len(generated) == np.sum(np.ceil(np.array(it.dataset.sizes) / step))
        for i in range(3):
            g = [elem[0] for elem in generated if elem[0] == i]
            assert len(g) == math.ceil(it.dataset.sizes[i] / step)


        step = 45
        x = AllShuffled()
        it = WindowedDatasetIterator(MockDatasetMedium(), window_size=5, step=step, shuffler=AllShuffled())
        q = [[e for e in x.iterator(it)] for i in range(250)]
        q1 = [len(b) for b in q]
        assert q1.count(q1[0]) == len(q1)


    def test_same_points(self):


        step = 5
        x = AllShuffled()
        it = WindowedDatasetIterator(MockDatasetBig(673), window_size=5, step=step)
        values1 = [e for e in x.iterator(it)]
        assert len(set(values1)) == len(values1)

        step = 5
        x = NotShuffled()
        it = WindowedDatasetIterator(MockDatasetBig(673), window_size=5, step=step)
        values2 = [e for e in x.iterator(it)]
        assert len(set(values2)) == len(values2)
        assert (sorted(values1, key=lambda x: (x[0], x[1])) == sorted(values2, key=lambda x: (x[0], x[1])))

        






import numpy as np
import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.iterators.batcher import Batcher
from ceruleo.transformation.features.scalers import MinMaxScaler
from ceruleo.transformation.features.selection import ByNameFeatureSelector
from ceruleo.transformation import ( Transformer)
import math

class MockDataset(AbstractPDMDataset):
    def __init__(self, nlives: int):

        self.lives = [
            pd.DataFrame({
                'feature1': np.linspace(0, (i+1)*100, 50),
                'feature2': np.linspace(-25, (i+1)*500, 50),
                'RUL': np.linspace(100, 0, 50)
            })
            for i in range(nlives-1)]

        self.lives.append(
            pd.DataFrame({
                'feature1': np.linspace(0, 5*100, 50),
                'feature2': np.linspace(-25, 5*500, 50),
                'feature3': np.linspace(-25, 5*500, 50),
                'RUL': np.linspace(100, 0, 50)
            })
        )

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return 'RUL'

    @property
    def n_time_series(self):
        return len(self.lives)


class TestBatcher():
    def test_batcher(self):
        features = ['feature1', 'feature2']
        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=['RUL'])
        transformer = Transformer(x, y)
        
        batch_size = 15
        window_size = 5
        ds = MockDataset(5)
        transformer.fit(ds)
        b = Batcher.new(ds.map(transformer), window_size, batch_size, 1)

        expected_size = math.ceil(((5*50) - (4*5)) / 15)
        assert len(b) == expected_size
        X, y, w = next(b)
        assert len(y.ravel()) == batch_size
        assert X.shape[0] == batch_size
        assert X.shape[1] == window_size
        assert X.shape[2] == 2


        features = ['feature1', 'feature2']
        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=['RUL'])
        transformer = Transformer(x, y)
        
        batch_size = 15
        window_size = 5
        ds = MockDataset(5)
        transformer.fit(ds)
        b = Batcher.new(ds.map(transformer), window_size, batch_size, 1, padding=True)

        expected_size = math.ceil(((5*50)) / 15)
        assert len(b) == expected_size
        X, y, w = next(b)
        assert len(y.ravel()) == batch_size
        assert X.shape[0] == batch_size
        assert X.shape[1] == window_size
        assert X.shape[2] == 2


import numpy as np
import pandas as pd
from ceruleo.transformation.features.denoising import (GaussianFilter,
    EWMAFilter, MeanFilter, MedianFilter, MultiDimensionalKMeans, OneDimensionalKMeans, SavitzkyGolayTransformer)


class TestDenoising:
    def test_SavitzkyGolayTransformer(self):

        remover = SavitzkyGolayTransformer(window=25)
        df = pd.DataFrame(
            {
                "a": np.random.rand(2000),
                "b": np.random.rand(2000),
            }
        )
        df_new = remover.fit_transform(df)
        
        assert df_new.shape[1] == 2 
        assert df_new['a'].max() <= df['a'].max()


    def test_MeanFilter(self):

        remover = MeanFilter(50, center=False)
        df = pd.DataFrame(
            {
                "a": np.random.rand(2000),
                "b": np.random.rand(2000),
            }
        )
        df_new = remover.fit_transform(df)
        
        assert df_new.shape[1] == 2 
        assert df_new['a'].max() <= df['a'].max()
        assert df_new['a'].min() >= df['a'].min()

    def test_MedianFilter(self):

        remover = MedianFilter(50)
        df = pd.DataFrame(
            {
                "a": np.random.rand(2000),
                "b": np.random.rand(2000),
            }
        )
        df_new = remover.fit_transform(df)
        
        assert df_new.shape[1] == 2 
        assert df_new['a'].max() <= df['a'].max()
        assert df_new['a'].min() >= df['a'].min()


    def test_OneDimensionalKMeans(self):
        remover = OneDimensionalKMeans(n_clusters=3)
        df = pd.DataFrame(
            {
                "a": np.random.rand(2000),
                "b": np.random.rand(2000),
            }
        )
        remover.partial_fit(df)
        df_new = remover.transform(df)
        
        assert df_new.shape[1] == 2 
        assert len(df_new['a'].unique()) == 3
        assert len(df_new['b'].unique()) == 3
        
    def test_MultiDimensionalKMeans(self):
        remover = MultiDimensionalKMeans(n_clusters=3)
        df = pd.DataFrame(
            {
                "a": np.random.rand(2000),
                "b": np.random.rand(2000),
            }
        )
        remover.partial_fit(df)
        df_new = remover.transform(df)
        
        assert df_new.shape[1] == 2 
        assert df_new[['a', 'b']].drop_duplicates().shape[0] == 3


    def test_EWMAFilter(self):
        remover = EWMAFilter(50)
        df = pd.DataFrame(
            {
                "a": np.random.rand(2000),
                "b": np.random.rand(2000),
            }
        )
        remover.partial_fit(df)
        df_new = remover.transform(df)
        
        assert df_new.shape[1] == 2 
        assert df_new['a'].max() <= df['a'].max()
        assert df_new['a'].min() >= df['a'].min()
        

    def test_GaussianFilter(self):
        remover = GaussianFilter(50, 5)
        df = pd.DataFrame(
            {
                "a": np.random.rand(2000),
                "b": np.random.rand(2000),
            }
        )
        remover.partial_fit(df)
        df_new = remover.transform(df)
        
        assert df_new.shape[1] == 2 
        assert df_new['a'].max() <= df['a'].max()
        assert df_new['a'].min() >= df['a'].min()
        


 


import numpy as np
from ceruleo.transformation.target import PicewiseRUL, RULBinarizer, TargetToClasses
import pandas as pd

class TestTargetTransformers():
    def test_PicewiseRUL(self):
        t = PicewiseRUL(max_life=26)
        d = np.linspace(0, 30, 50)
        d = pd.DataFrame({'RUL': d})
        d1 = t.fit_transform(d)
        assert d1.max().RUL == 26

    def test_classes(self):
        t = TargetToClasses(bins=[15, 50, 150, 200])
        d = pd.DataFrame({'RUL': np.linspace(0, 500, 50)})
        d1 = t.fit_transform(d)

        assert np.all(((d <=15 ).values == (d1 == 0).values))
        assert np.all(((((d >15)  & (d <= 50))).values == (d1 == 1).values))

    def test_binarizer(self):
        t = RULBinarizer(50)
        d = pd.DataFrame({'RUL': np.linspace(0, 500, 50)})
        d1 = t.fit_transform(d)
        assert np.all(((d > 50 ).values == (d1 == 1).values))


import numpy as np
import pandas as pd
from pyexpat import features
from scipy.stats import entropy
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.transformation import Concatenate as TransformationConcatenate
from ceruleo.transformation import Transformer
from ceruleo.transformation.features.imputers import PerColumnImputer
from ceruleo.transformation.features.outliers import IQROutlierRemover
from ceruleo.transformation.features.scalers import MinMaxScaler
from ceruleo.transformation.features.selection import ByNameFeatureSelector
from ceruleo.transformation.features.split import SplitByCategory
from ceruleo.transformation.features.transformation import MeanCentering
from ceruleo.transformation.functional.concatenate import Concatenate
from ceruleo.transformation.functional.graph_utils import root_nodes
from ceruleo.transformation.functional.pipeline.pipeline import make_pipeline


def gaussian(N: int, mean: float = 50, std: float = 10):
    return np.random.randn(N) * std + mean


class MockDatasetCategorical(AbstractPDMDataset):
    def build_df(self):
        N = 50
        return pd.DataFrame(
            {
                "Categorical": ["a"] * N + ["b"] * N,
                "feature1": np.hstack(
                    (gaussian(N, self.mean_a_f1), gaussian(N, self.mean_b_f1))
                ),
                "feature2": np.hstack(
                    (gaussian(N, self.mean_a_f2), gaussian(N, self.mean_b_f2))
                ),
            }
        )

    def __init__(self, N: int = 5):
        super().__init__()
        self.mean_a_f1 = 50
        self.mean_b_f1 = -16

        self.mean_a_f2 = 90
        self.mean_b_f2 = 250

        self.lives = [self.build_df() for i in range(N)]
        life_4 = self.lives[4]
        life_4.loc[life_4.index[50], "feature1"] = 591212
        life_4.loc[life_4.index[21], "feature2"] = 591212

        life_3 = self.lives[3]
        life_3.loc[life_4.index[88], "feature1"] = 591212
        life_3.loc[life_4.index[25], "feature2"] = 591212

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class MockDataset(AbstractPDMDataset):
    def __init__(self):
        super().__init__()
        self.lives = [
            pd.DataFrame({"a": [1, 2, 3, 4], "b": [2, 4, 6, 8], "RUL": [4, 3, 2, 1]}),
            pd.DataFrame(
                {"a": [150, 5, 14, 24], "b": [-52, -14, -36, 8], "RUL": [4, 3, 2, 1]}
            ),
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class MockDataset1(AbstractPDMDataset):
    def __init__(self):
        super().__init__()
        self.lives = [
            pd.DataFrame({"a": [1, 2, 3, 4], "b": [1, 2, 3, 4], "RUL": [4, 3, 2, 1]}),
            pd.DataFrame({"a": [2, 4, 6, 8], "b": [2, 4, 6, 8], "RUL": [4, 3, 2, 1]}),
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class MockDataset2(AbstractPDMDataset):
    def __init__(self, n: int = 5):
        super().__init__()
        self.lives = [
            pd.DataFrame(
                {"a": ["A", "A", "A", "A"], "b": [1, 2, 3, 4], "RUL": [4, 3, 2, 1]}
            )
            for i in range(n)
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class TestPipeline:
    def test_FitOrder(self):

        dataset = MockDataset()

        pipe = ByNameFeatureSelector(features=["a", "b"])
        pipe = MeanCentering()(pipe)
        pipe = MinMaxScaler(range=(-1, 1), name="Scaler")(pipe)

        target_pipe = ByNameFeatureSelector(features=["RUL"])

        test_transformer = Transformer(pipelineX=pipe, pipelineY=target_pipe)

        test_transformer.fit(dataset)

        X, y, sw = test_transformer.transform(dataset[0])

        assert X.shape[1] == 2
        df_dataset = dataset.to_pandas()

        centered_df = df_dataset[["a", "b"]] - df_dataset[["a", "b"]].mean()
        scaler = test_transformer.pipelineX.find_node("Scaler")
        assert scaler.data_min.equals(centered_df.min(axis=0))
        assert scaler.data_max.equals(centered_df.max(axis=0))

    def test_FitOrder2(self):
        dataset = MockDataset()

        pipe_a = ByNameFeatureSelector(features=["a"])
        pipe_a = MeanCentering()(pipe_a)
        scaler_a = MinMaxScaler(range=(-1, 1), name="a")
        pipe_a = scaler_a(pipe_a)

        pipe_b = ByNameFeatureSelector(features=["b"])
        pipe_b = MeanCentering()(pipe_b)
        scaler_b = MinMaxScaler(range=(-1, 1), name="b")
        pipe_b = scaler_b(pipe_b)

        pipe = TransformationConcatenate()([pipe_a, pipe_b])

        target_pipe = ByNameFeatureSelector(features=["RUL"])

        test_transformer = Transformer(pipelineX=pipe, pipelineY=target_pipe)

        test_transformer.fit(dataset)

        X, y, sw = test_transformer.transform(dataset[0])

        assert X.shape[1] == 2
        df_dataset = dataset.to_pandas()
        centered_df = df_dataset[["a", "b"]] - df_dataset[["a", "b"]].mean()

        assert scaler_a.data_min.equals(centered_df.min(axis=0)[["a"]])
        assert scaler_b.data_max.equals(centered_df.max(axis=0)[["b"]])

    def test_Concatenate(self):
        dataset = MockDataset1()

        pipe = ByNameFeatureSelector(features=["a"])
        pipe = MinMaxScaler(range=(-1, 1))(pipe)

        pipe2 = ByNameFeatureSelector(features=["b"])
        pipe2 = MinMaxScaler(range=(-5, 0))(pipe2)

        pipe = TransformationConcatenate()([pipe, pipe2])
        pipe = MeanCentering()(pipe)

        target_pipe = ByNameFeatureSelector(features=["RUL"])

        test_transformer = Transformer(pipelineX=pipe, pipelineY=target_pipe)

        test_transformer.fit(dataset)

        df = dataset.to_pandas()[["a", "b"]]

        data_min = df.min()
        data_max = df.max()

        gt = (df - data_min) / (data_max - data_min)
        gt["a"] = gt["a"] * (1 - (-1)) + (-1)
        gt["b"] = gt["b"] * (0 - (-5)) + (-5)
        gt = gt - gt.mean()

        X, y, sw = test_transformer.transform(dataset[0])

        assert (np.mean((gt.iloc[:4, :].values - X.values) ** 2)) < 0.0001
        X, y, sw = test_transformer.transform(dataset[1])
        assert (np.mean((gt.iloc[4:, :].values - X.values) ** 2)) < 0.0001

        assert X.shape[1] == 2

    def test_subpipeline(self):
        dataset = MockDatasetCategorical()
        pipe = ByNameFeatureSelector(features=["Categorical", "feature1", "feature2"])
        bb = make_pipeline(
            IQROutlierRemover(lower_quantile=0.05, upper_quantile=0.95, clip=True),
            MinMaxScaler(range=(-1, 1)),
            PerColumnImputer(),
        )
        pipe = SplitByCategory(features="Categorical", pipeline=bb)(pipe)

        target_pipe = ByNameFeatureSelector(features=["RUL"])

        test_transformer = Transformer(pipelineX=pipe)
        test_transformer.fit(dataset)

        q = np.hstack([d[d["Categorical"] == "a"]["feature1"] for d in dataset])
        approx_cat_a_feature1_1_quantile = np.quantile(q, 0.05)
        approx_cat_a_feature1_3_quantile = np.quantile(q, 0.95)
        r = root_nodes(pipe)[0]
        IQR_Node = r.next[0].next[1].next[0]
        real_cat_a_feature1_1_quantile = IQR_Node.Q1["feature1"]
        real_cat_a_feature1_3_quantile = IQR_Node.Q3["feature1"]

        assert approx_cat_a_feature1_1_quantile - real_cat_a_feature1_1_quantile < 5
        assert approx_cat_a_feature1_3_quantile - real_cat_a_feature1_3_quantile < 5

        assert test_transformer.transform(dataset[4])[0]["feature1"].iloc[50] - 1 < 0.01
        assert test_transformer.transform(dataset[4])[0]["feature2"].iloc[21] - 1 < 0.01

        d = dataset[4]
        aa = d[d["Categorical"] == "a"]["feature1"]
        counts_before_transformation, _ = np.histogram(aa)
        counts_before_transformation = counts_before_transformation / np.sum(
            counts_before_transformation
        )

        bb = test_transformer.transform(dataset[4])[0]["feature1"]
        counts_after_transformation, _ = np.histogram(bb[:50])
        counts_after_transformation = counts_after_transformation / np.sum(
            counts_after_transformation
        )
        assert entropy(counts_before_transformation, counts_after_transformation) < 0.01

    def test_split_one_category(self):
        dataset_orig = MockDataset2()
        dataset = dataset_orig[0:5]

        pipe = ByNameFeatureSelector(features=["a", "b"])
        scaler_pipe = make_pipeline(MinMaxScaler(range=(-1, 1), name="Scaler"))
        pipe = SplitByCategory(
            features="a", pipeline=scaler_pipe, add_default=False
        )(pipe)
        pipe = MeanCentering()(pipe)

        pipe = MinMaxScaler(range=(-1, 1), name="Scaler2")(pipe)

        target_pipe = ByNameFeatureSelector(features=["RUL"])

        test_transformer = Transformer(pipelineX=pipe, pipelineY=target_pipe)

        test_transformer.fit(dataset)

        X, y, sw = test_transformer.transform(dataset[0])

        assert X.shape[1] == 1


import itertools
import logging
from typing import List, Optional

import numpy as np
import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.dataset.utils import iterate_over_features
from scipy.special import kl_div
from scipy.stats import wasserstein_distance
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)


def histogram_per_cycle(
    cycle: pd.DataFrame,
    feature: str,
    bins_to_use: np.ndarray,
    normalize: bool = True,
) -> List[np.ndarray]:
    """Compute the histogram of a feature in a run-to-failure cycle

    Args:
        cycle (pd.DataFrame): The run-to-failure cycle
        feature (str): The  feature to compute the histogram
        bins_to_use (np.ndarray): Number of bins to use
        normalize (bool, optional): Wheter to normalize the histogram. Defaults to True.

    Returns:
        List[np.ndarray]: The histogram of the feature
    """
    try:
        d = cycle[feature]
        h, _ = np.histogram(d, bins=bins_to_use)

        if normalize:
            h = h / np.sum(h)
            h += 1e-50
        return h
    except Exception as e:
        logger.info(f"Error {e} when computing the distribution for feature {feature}")


def compute_bins(ds: AbstractPDMDataset, feature: str, number_of_bins: int = 15):
    min_value = ds.get_features_of_life(0)[feature].min()
    max_value = ds.get_features_of_life(0)[feature].max()

    for life in iterate_over_features(ds):
        min_value = min(np.min(life[feature]), min_value)
        max_value = max(np.max(life[feature]), max_value)
    bins_to_use = np.linspace(min_value, max_value, number_of_bins + 1)
    return bins_to_use


def features_divergeces(
    ds: AbstractPDMDataset,
    number_of_bins: int = 15,
    columns: Optional[List[str]] = None,
    show_progress: bool = False,
) -> pd.DataFrame:
    """
    Compute the divergence between features

    Parameters:
        ds: The dataset
        number_of_bins: Number of bins
        columns: Which columns to use

    Returns:
        A DataFrame in which each row contains the distances between a feature of two run-to-failure cycle with the following columns:

            - Cycle 1: Run-to-failure cycle 1
            - Cycle 2: Run-to-failure cycle 2
            - Wasserstein: Wasserstein
            - KL: KL Divergence
            - feature: The feature name
    """
    if columns is None:
        columns = ds.numeric_features()

    features_bins = {}
    iterator = tqdm(columns) if show_progress else columns

    for feature in iterator:
        features_bins[feature] = compute_bins(ds, feature, number_of_bins)

    histograms = {}
    for life in iterate_over_features(ds):
        for feature in columns:
            if feature not in histograms:
                histograms[feature] = []
            histograms[feature].append(
                histogram_per_cycle(life, feature, features_bins[feature])
            )

    df_data = []
    for feature in columns:
        data = {}
        for (i, h1), (j, h2) in itertools.combinations(
            enumerate(histograms[feature]), 2
        ):
            kl = (np.mean(kl_div(h1, h2)) + np.mean(kl_div(h2, h1))) / 2
            wd = wasserstein_distance(h1, h2)
            df_data.append(
                (
                    i,
                    j,
                    ds.get_features_of_life(i).shape[0],
                    ds.get_features_of_life(j).shape[0],
                    abs(ds.get_features_of_life(i).shape[0]-ds.get_features_of_life(j).shape[0]),
                    wd,
                    kl,
                    feature,
                )
            )
    df = pd.DataFrame(
        df_data,
        columns=[
            "Cycle 1",
            "Cycle 2",
            "Cycle 1 length",
            "Cycle 2 length",
            "Abs Length difference",           
            "Wasserstein",
            "KL",
            "feature",
        ],
    )

    return df



import uuid
from copy import deepcopy
from typing import List, Optional, Union

from sklearn.base import BaseEstimator


class TransformerStepMixin(BaseEstimator):
    name: Optional[str]
    previous: List["TransformerStepMixin"]
    next: List["TransformerStepMixin"]
    uuid: str
    prefer_partial_fit: bool

    def __init__(self, *, name: Optional[str] = None, prefer_partial_fit:bool =True):
        self.name_ = name
        self.previous = []
        self.next = []
        self.uuid = "".join(str(uuid.uuid4()).split("-"))
        self.prefer_partial_fit = prefer_partial_fit


    @property
    def name(self):
        if self.name_ is not None:
            return self.name_
        else:
            return self.__class__.__name__

    @name.setter
    def name(self, value: str):
        self.name_ = value

    def __call__(
        self, prev: Union[List["TransformerStepMixin"], "TransformerStepMixin"]
    ):
        self._add_previous(prev)
        return self

    def _add_previous(
        self, prev: Union[List["TransformerStepMixin"], "TransformerStepMixin"]
    ):
        if not isinstance(prev, list):
            prev = [prev]
        for p in prev:
            p.add_next(self)

    def remove_previous(self, node):
        self.previous.remove(node)

    def disconnect(self, node):
        if node in self.previous:
            self.previous.remove(node)
            node.disconnect(self)
        elif node in self.next:
            self.next.remove(node)
            node.disconnect(self)

    def add_next(
        self, nodes: Union[List["TransformerStepMixin"], "TransformerStepMixin"]
    ):
        if not isinstance(nodes, list):
            nodes = [nodes]
        for n in nodes:
            self.next.append(n)
            n.previous.append(self)

    def description(self):
        return self.name

    def __str__(self):
        return self.name

    
    def clear_connections(self):
        self.previous = []
        self.next = []


from typing import Optional, Union

import numpy as np
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.dataset.utils import iterate_over_target
from ceruleo.results.results import FittedLife

from ceruleo.dataset.transformed import TransformedDataset


class BaselineModel:
    """
    Predict the RUL using the mean or the median value of the duration of the dataset

    Parameters:
        mode: Method for computing the duration of the dataset. Possible values are: 'mean' and 'median'
    """

    def __init__(self, mode: str = "mean", RUL_threshold: Optional[float] = None):
        self.mode = mode
        self.RUL_threshold = RUL_threshold

    def fit(self, ds: Union[TransformedDataset, AbstractPDMDataset]):
        """Compute the mean or median RUL using the given dataset

        Parameters:
            ds:  Dataset from which obtain the true RUL
        """
        true = []
        for y in iterate_over_target(ds):
            y = y
            degrading_start, time = FittedLife.compute_time_feature(
                y, self.RUL_threshold
            )

            true.append(y.iloc[0] + time[degrading_start])

        if self.mode == "mean":
            self.fitted_RUL = np.mean(true)
        elif self.mode == "median":
            self.fitted_RUL = np.median(true)

    def predict(self, ds: TransformedDataset) -> np.ndarray:
        """
        Predict the whole life using the fitted values

        Parameters:
            ds: Dataset iterator from which obtain the true RUL

        Returns:
            Predicted RUL
        """
        output = []
        for y in iterate_over_target(ds):
            _, time = FittedLife.compute_time_feature(y, self.RUL_threshold)
            y_pred = np.clip(self.fitted_RUL - time, 0, self.fitted_RUL)
            output.append(y_pred)
        return np.concatenate(output)


class FixedValueBaselineModel:
    """
    A model that predicts always  the same duration for each run-to-failure cycle

    Parameters:
        value: Fixed RUL
    """

    def __init__(self, *, value: float):
        self.value = value

    def fit(self, *args):
        return self

    def predict(
        self, ds: TransformedDataset, RUL_threshold: Optional[float] = None
    ) -> np.ndarray:
        """
        Predict the whole life using the fixed values

        Parameters:
            ds: Dataset iterator from which obtain the true RUL

        Returns:
            Predicted RUL
        """
        output = []
        for y in iterate_over_target(ds):
            _, time = FittedLife.compute_time_feature(y, RUL_threshold)
            y_pred = np.clip(self.value - time, 0, self.value)
            output.append(y_pred)
        return np.concatenate(output)



import numpy as np


class LRUDataCache:
    def __init__(self, max_elem):
        self.max_elem = max_elem
        self.data = {}

    def get(self, key):
        self.data[key]['hit'] += 1

        return self.data[key]['elem']

    def add(self, key, elem):
        if len(self.data) == self.max_elem:
            keys = list(self.data.keys())
            key_to_remove = np.argmin([self.data[k]['hit'] for k in keys])
            del self.data[keys[key_to_remove]]
        self.data[key] = {
            'elem':  elem,
            'hit': 0
        }
        return elem

    def __len__(self):
        return len(self.data)



from pathlib import Path
from typing import List, Optional, Union

import pandas as pd
from ceruleo import CACHE_PATH
from ceruleo.transformation.functional.pipeline.cache_store import (
    CacheStoreType, GraphTraversalCacheMemory, GraphTraversalCacheShelveStore)
from ceruleo.transformation.functional.pipeline.utils import (decode_tuple,
                                                              encode_tuple)
from ceruleo.transformation.functional.transformerstep import TransformerStep


class CachedGraphTraversal:
    """Iterator for a graph nodes. 


    The cache data structures has the following form
    Current Node -> Previous Nodes -> [Transformed Dataset]

    * cache[n]:
        contains a dict with one key for each previous node
    * cache[n][n.previous[0]]
        A list with each element of the dataset transformed in
        up to n.previous[0]


    Parameters:
        
        root_nodes: Initial nodes of the graph
        dataset: Each node visit the dataset
        cache_path: Where to store the cache
        cache_type: Mode for storing the intermediate steps

    """
    def __init__(
        self,
        root_nodes,
        dataset,
        cache_path: Optional[Path] = CACHE_PATH,
        cache_type: CacheStoreType = CacheStoreType.SHELVE,
    ):
        if cache_type == CacheStoreType.SHELVE:
            self.transformed_cache = GraphTraversalCacheShelveStore(cache_path)
        elif cache_type == CacheStoreType.MEMORY:
            self.transformed_cache = GraphTraversalCacheMemory()

        for r in root_nodes:
            for i, df in enumerate(dataset):
                self.transformed_cache[encode_tuple((r, None, i))] = df

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.transformed_cache.close()

    def clear_cache(self):
        self.transformed_cache.claer()

    def state_up_to(self, current_node: TransformerStep, dataset_element: int):

        previous_node = current_node.previous

        if len(previous_node) > 1:
            return [
                self.transformed_cache[encode_tuple((current_node, p, dataset_element))]
                for p in previous_node
            ]
        else:
            if len(previous_node) == 1:
                previous_node = previous_node[0]
            else:
                previous_node = None

            return self.transformed_cache[
                encode_tuple((current_node, previous_node, dataset_element))
            ]

    def clean_state_up_to(self, current_node: TransformerStep, dataset_element: int):

        previous_node = current_node.previous
        for p in previous_node:
            self.transformed_cache[
                encode_tuple((current_node, p, dataset_element))
            ] = None

    def store(
        self,
        next_node: Optional[TransformerStep],
        node: TransformerStep,
        dataset_element: int,
        new_element: pd.DataFrame,
    ):
        self.transformed_cache[
            encode_tuple((next_node, node, dataset_element))
        ] = new_element

    def remove_state(self, nodes: Union[TransformerStep, List[TransformerStep]]):
        if not isinstance(nodes, list):
            nodes = [nodes]
        for n in nodes:
            keys_to_remove = self.get_keys_of(n)
            for k in keys_to_remove:
                self.transformed_cache.pop(k)

    def get_keys_of(self, n):
        return [
            k
            for k in self.transformed_cache.keys()
            if decode_tuple(k)[0] == str(hash(n))
        ]



from typing import Tuple
import tensorflow as tf

from ceruleo.models.keras.layers import ExpandDimension
from tensorflow.keras import Input, Model
from tensorflow.keras.layers import (
    Activation,
    Add,
    BatchNormalization,
    Conv2D,
    Dense,
    Dropout,
    Flatten,
)


def MultiScaleConvolutionalModel(
    input_shape: Tuple[int, int],
    *,
    n_msblocks: int,
    scales: list,
    n_hidden: int,
    l2: float = 0.5,
    dropout: float = 0.5,
    activation: str = 'relu'
):
    """Remaining useful life prediction using multi-scale deep convolutional neural network


    Li, H., Zhao, W., Zhang, Y., & Zio, E. (2020). 
    Remaining useful life prediction using multi-scale deep convolutional neural network. 
    Applied Soft Computing, 89, 106113.
    
    (Reference)[https://www.sciencedirect.com/science/article/pii/S1568494620300533?casa_token=wp27UPxwVTIAAAAA:e_fMuKsvfQf8VZ7DgsXKCi6mpnrcx2hI0tbfe5xrLRhxmc2vaR-uW_Qq23v5yBqziBoSM5gu]
    
    Parameters:

        input_shape: Input shape of the iterator
        n_msblocks: Number of scale blocks
        scales: List of scales
        n_hidden: Hidden convolutional dimensions
        l2: regularization
        dropout: Dropout factor
        activation
    """

    def _create_scale_conv(hidden, scale):
        return Conv2D(hidden, (scale, 1), padding="same", activation=activation)

    input = Input(input_shape)
    x = input
    x = ExpandDimension()(x)
    for i in range(n_msblocks):
        convs = [
            _create_scale_conv(n_hidden, scale)(x) for scale in scales
        ]
        x = Add()(convs)
    x = Conv2D(10, (10, 1), padding="same", activation=activation)(x)
    x = Conv2D(1, (3, 1), padding="same", activation=activation)(x)
    x = Flatten()(x)
    x = Dense(64, kernel_regularizer=tf.keras.regularizers.l2(l2))(x)
    x = BatchNormalization()(x)
    x = Activation("tanh")(x)
    x = Dropout(dropout)(x)
    x = Dense(1, activation='relu')(x)
    return Model(inputs=[input], outputs=[x])



import torch
from ceruleo.iterators.iterators import WindowedDatasetIterator
import math 

class IterableDataset(torch.utils.data.IterableDataset):
     def __init__(self, iterator:WindowedDatasetIterator):
         super(IterableDataset).__init__()
         self.iterator = iterator

     def __iter__(self):
         worker_info = torch.utils.data.get_worker_info()
         if worker_info is None:  # single-process data loading, return the full iterator
             iter_start = self.start
             iter_end = self.end
         else:  # in a worker process
             # split workload
             per_worker = int(math.ceil((self.end - self.start) / float(worker_info.num_workers)))
             worker_id = worker_info.id
             iter_start = self.start + worker_id * per_worker
             iter_end = min(iter_start + per_worker, self.end)
         return iter(range(iter_start, iter_end))


import itertools
import logging
from typing import Dict, List, Optional

import emd
import mmh3
import numpy as np
import pandas as pd
from numpy.lib.arraysetops import isin
from pandas.core.window.expanding import Expanding
from pyts.transformation import ROCKET as pyts_ROCKET
from functools import reduce
import pywt

# try:
#    from temporis.transformation.features.hurst import hurst_exponent
# except:
#    pass
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.rolling_windows import apply_rolling_data
from ceruleo.transformation.functional.transformers import Transformer
from ceruleo.transformation.utils import SKLearnTransformerWrapper
from pandas.core.window.rolling import Rolling


logger = logging.getLogger(__name__)


class TimeToPreviousBinaryValue(TransformerStep):
    """Return a column with increasing number"""

    def time_to_previous_event(self, X: pd.DataFrame, c: str):
        def min_idex(group):
            if group.iloc[0, 0] == 0:
                return np.nan
            else:
                return np.min(group.index)

        X_c_cumsum = X[[c]].cumsum()
        min_index = X_c_cumsum.groupby(c).apply(min_idex)
        X_merged = X_c_cumsum.merge(
            pd.DataFrame(min_index, columns=["start"]), left_on=c, right_index=True
        )
        return X_merged.index - X_merged["start"]

    def transform(self, X: pd.DataFrame):
        new_X = pd.DataFrame(index=X.index)
        for c in X.columns:
            new_X[f"ttp_{c}"] = self.time_to_previous_event(X, c)
        return new_X


class ColumnWiseSum(TransformerStep):
    """
    Compute the column-wise sum each column

    Parameters:
        column_name: Name of the unique column which is returned 
    """

    def __init__(self, column_name: str, name: Optional[str] = None):
        super().__init__(name=name)
        self.column_name = column_name

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life 

        Parameters:
            X: The input life 

        Returns:
            A single-column DataFrame containing the column-wise sum for each input sample
        """
        return pd.DataFrame(X.sum(axis=1), columns=[self.column_name])


class SampleNumber(TransformerStep):
    """Return a column with increasing number"""

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life

        Parameters:
            X: The input life

        Returns:
            A DataFrame with increasing sample indexes. 
        """
        df = pd.DataFrame(index=X.index)
        df["sample_number"] = list(range(X.shape[0]))
        return df


class OneHotCategorical(TransformerStep):
    """Compute a one-hot encoding for a given feature

    Parameters:
        feature: Feature name from which compute the one-hot encoding
        name: Step name, by default None
    """

    def __init__(
        self,
        *,
        feature: Optional[str] = None,
        categories: Optional[List[any]] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.feature = feature
        self.categories = categories
        self.fixed_categories = True
        if self.categories is None:
            self.categories = set()
            self.fixed_categories = False
        self.encoder = None

    def partial_fit(self, X: pd.DataFrame, y=None):
        if self.fixed_categories:
            return self
        if self.feature is None:
            self.feature = X.columns[0]
        self.categories.update(set(X[self.feature].unique()))
        return self

    def fit(self, X: pd.DataFrame, y=None):
        if self.fixed_categories:
            return self
        if self.feature is None:
            self.feature = X.columns[0]
        self.categories.update(set(X[self.feature].unique()))
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Apply the transformation to the input life

        Parameters:
            X: The input life

        Returns:
            A DataFrame with shape equal to (X.shape[0],n_unique(feature)) containin the One Hot Encoding for the input feature
        """
        categories = sorted(list([c for c in self.categories if c is not None]))
        d = pd.Categorical(X[self.feature], categories=categories)

        df = pd.get_dummies(d)
        df.index = X.index
        return df


class HashingEncodingCategorical(TransformerStep):
    """Compute a simple numerical encoding for a given feature

    Parameters:
        nbins: Number of bins after the hash
        feature: Feature name from which compute the simple encoding
        name: Step name
    """

    def __init__(
        self, *, nbins: int, feature: Optional[str] = None, name: Optional[str] = None
    ):
        super().__init__(name=name)
        self.nbins = nbins
        self.feature = feature
        self.categories = set()
        self.encoder = None

    def transform(self, X: pd.DataFrame, y: Optional[type]=None) -> pd.DataFrame:
        """
        Return a new DataFrame with the feature  encoded with integer numbers

        Parameters;
            X: The input life
            y: [type], optional

        Parameters:
            X: The input life
        
        Returns:
            A new dataframe with the same index as the input with 1 column containing the encoding of the input feature. 
        """

        def hash(x):
            if isinstance(x, int):
                x = x.to_bytes((x.bit_length() + 7) // 8, "little")
            return (mmh3.hash(x) & 0xFFFFFFFF) % self.nbins

        if self.feature is None:
            self.feature = X.columns[0]
        X_new = pd.DataFrame(index=X.index)
        X_new["encoding"] = X[self.feature].map(hash)
        return X_new


class SimpleEncodingCategorical(TransformerStep):
    """Compute a simple numerical encoding for a given feature

    Parameters:
        feature: Feature name from which compute the simple encoding
        name: Step name, by default None
    """

    def __init__(self, *, feature: Optional[str] = None, name: Optional[str] = None):
        super().__init__(name=name)
        self.feature = feature
        self.categories = set()
        self.encoder = None

    def partial_fit(self, X: pd.DataFrame, y=None) -> "SimpleEncodingCategorical":
        """Compute incrementally the set of possible categories

        Parameters:
            X: The input life

        Returns:
            Instance of class SimpleEncodingCategorical
        """
        if self.feature is None:
            self.feature = X.columns[0]

        self.categories.update(set(X[self.feature].unique()))

        return self

    def fit(self, X: pd.DataFrame, y=None) -> "SimpleEncodingCategorical":
        """
        Compute the set of possible categories

        Parameters:
            X: The input life
        
        Returns:
            Instance of class SimpleEncodingCategorical
        """
        if self.feature is None:
            self.feature = X.columns[0]

        self.categories.update(set(X[self.feature].unique()))
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """Return a new DataFrame with the feature  encoded with integer numbers

        Parameters:
            X: The input life

        Returns:
            A new dataframe with the same index as the input with 1 column with the Simple Encoding of the input feature. 
        """
        categories = sorted(list([c for c in self.categories if c is not None]))
        d = pd.Categorical(X[self.feature], categories=categories)
        return pd.DataFrame({"encoding": d.codes}, index=X.index)


def rolling_kurtosis(s: pd.Series, window, min_periods):
    return s.rolling(window, min_periods=min_periods).kurt(numeric_only=True)


class LifeStatistics(TransformerStep):
    """Compute diverse number of features for each life.

    Returns a 1 row with the statistics computed for every feature


    The possible features are:

    - Kurtosis
    - Skewness
    - Max
    - Min
    - Std
    - Peak
    - Impulse
    - Clearance
    - RMS
    - Shape
    - Crest
    - Hurst


    Parameters:
        to_compute: List of the features to compute, by default None. Valid values are:'kurtosis', 'skewness', 'max', 'min', 'std', 'peak', 'impulse','clearance', 'rms', 'shape', 'crest', 'hurst'
        name: Name of the step, by default None
    """

    def __init__(
        self, *, to_compute: Optional[List[str]] = None, name: Optional[str] = None
    ):
        super().__init__(name=name)
        valid_stats = [
            "kurtosis",
            "skewness",
            "max",
            "min",
            "std",
            "peak",
            "impulse",
            "clearance",
            "rms",
            "shape",
            "crest",
        ]
        if to_compute is None:
            self.to_compute = valid_stats
        else:
            for f in to_compute:
                if f not in valid_stats:
                    raise ValueError(
                        f"Invalid feature to compute {f}. Valids are {valid_stats}"
                    )
            self.to_compute = to_compute

    def partial_fit(self, X, y=None):
        return self

    def fit(self, X, y=None):
        return self

    def _kurtosis(self, s: pd.Series):
        return s.kurt(skipna=True)

    def _skewness(self, s: pd.Series):
        return s.skew(skipna=True)

    def _max(self, s: pd.Series):
        return s.max(skipna=True)

    def _min(self, s: pd.Series):
        return s.min(skipna=True)

    def _std(self, s: pd.Series):
        return s.std(skipna=True)

    def _peak(self, s: pd.Series):
        return s.max(skipna=True) - s.min(skipna=True)

    def _impulse(self, s: pd.Series):
        m = s.abs().mean()
        if m > 0:
            return self._peak(s) / m
        else:
            return 0

    def _clearance(self, s: pd.Series):
        m = s.abs().pow(1.0 / 2).mean()
        if m > 0:
            return (self._peak(s) / m) ** 2
        else:
            return 0

    def _rms(self, s: pd.Series):
        return np.sqrt(s.pow(2).mean(skipna=True))

    def _shape(self, s: pd.Series):
        m = s.abs().mean(skipna=True)
        if m > 0:
            return self._rms(s) / m
        else:
            return 0

    def _crest(self, s: pd.Series):
        m = self._rms(s)
        if m > 0:
            return self._peak(s) / m
        else:
            return 0

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Compute features from the given life

        Parameters:
            X: The input life

        Returns:
            A new DataFrame with one row and with n columns. Let m be the number of features of the life and f the len(to_compute), then n = m x f,
        """
        X_new = pd.DataFrame(index=[0])
        for c in X.columns:
            for stats in self.to_compute:
                X_new[f"{c}_{stats}"] = getattr(self, f"_{stats}")(X[c])
        return X_new


class RollingStatistics(TransformerStep):
    """Compute diverse number of features using an rolling window.

    For each feature present in the life a number of feature will be computed for each time stamp

    The possible features are:

    Time domain:

    - Kurtosis
    - Skewness
    - Max
    - Min
    - Std
    - Peak
    - Impulse
    - Clearance
    - RMS
    - Shape
    - Crest

    Parameters:
        window: Size of the rolling window, by default 15
        min_points: The minimun number of points of the expanding window
        to_compute: Name of features to compute. Possible values are: 'kurtosis', 'skewness', 'max', 'min', 'std', 'peak', 'impulse', 'clearance', 'rms', 'shape', 'crest'
        name: Name of the step, by default None

    """

    def __init__(
        self,
        *,
        window: int = 15,
        min_points=2,
        to_compute: Optional[List[str]] = None,
        specific: Optional[Dict[str, List[str]]] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.window = window
        self.min_points = min_points
        valid_stats = [
            "mean",
            "kurtosis",
            "skewness",
            "max",
            "min",
            "std",
            "peak",
            "impulse",
            "clearance",
            "rms",
            "shape",
            "crest",
            "deviance",
            "std_atan",
            "std_acosh",
            "std_asinh",
            "energy",
        ]

        if to_compute is not None and specific is not None:
            raise ValueError("Only one of to_compute or specific should be used")
        self.specific = specific
        self.to_compute = to_compute
        if to_compute is None:
            if specific is None:
                self.to_compute = valid_stats
            else:
                self.specific = specific
        else:
            for f in to_compute:
                if f not in valid_stats:
                    raise ValueError(
                        f"Invalid feature to compute {f}. Valids are {valid_stats}"
                    )
            self.to_compute = to_compute

    def partial_fit(self, X, y=None):
        return self

    def fit(self, X, y=None):
        return self

    def _std_asinh(self, X, rolling, abs_rolling):
        return (
            X.apply(np.arcsinh).rolling(self.window, self.min_points).std(numeric_only=True)
        )

    def _std_acosh(self, X, rolling, abs_rolling):
        return (
            X.apply(np.arccosh).rolling(self.window, self.min_points).std(numeric_only=True)
        )

    def _energy(self, X, rolling, abs_rolling):
        return X.pow(2).rolling(self.window, self.min_points).sum()

    def _std_atan(self, X, rolling, abs_rolling):
        return (
            X.apply(np.arctan)
            .rolling(self.window, self.min_points)
            .std(numeric_only=True)
        )

    def _mean(self, X, rolling: Rolling, abs_rolling: Rolling):
        return rolling.mean(numeric_only=True)

    def _kurtosis(self, X, rolling: Rolling, abs_rolling: Rolling):
        return rolling.kurt(numeric_only=True)

    def _skewness(self, X, rolling: Rolling, abs_rolling: Rolling):
        return rolling.skew(numeric_only=True)

    def _max(self, X, rolling: Rolling, abs_rolling: Rolling):
        return rolling.max(numeric_only=True)

    def _min(self, X, rolling: Rolling, abs_rolling: Rolling):
        return rolling.min(numeric_only=True)

    def _std(self, X, rolling: Rolling, abs_rolling: Rolling):
        return rolling.std(numeric_only=True)

    def _peak(self, X, rolling: Rolling, abs_rolling: Rolling):
        return rolling.max(numeric_only=True) - rolling.min(numeric_only=True)

    def _impulse(self, X, rolling: Rolling, abs_rolling: Rolling):
        return self._peak(X, rolling, abs_rolling) / abs_rolling.mean()

    def _deviance(self, X, rolling: Rolling, abs_rolling: Rolling):
        return (X - rolling.mean()) / rolling.std()

    def _clearance(self, X, rolling: Rolling, abs_rolling: Rolling):
        return self._peak(X, rolling, abs_rolling) / X.abs().pow(1.0 / 2).rolling(
            self.window, self.min_points
        ).mean().pow(2)

    def _rms(self, X, rolling: Rolling, abs_rolling: Rolling):
        return (
            X.pow(2)
            .rolling(self.window, self.min_points)
            .mean(numeric_only=True)
            .pow(1 / 2.0)
        )

    def _shape(self, X, rolling: Rolling, abs_rolling: Rolling):
        return self._rms(X, rolling, abs_rolling) / abs_rolling.mean(numeric_only=True)

    def _crest(self, X, rolling, abs_rolling):
        return self._peak(X, rolling, abs_rolling) / self._rms(X, rolling, abs_rolling)

    def _compute_column_names(self, X: pd.DataFrame):
        columns = []
        if self.to_compute is not None:
            for stats in self.to_compute:
                for c in X.columns:
                    columns.append(f"{c}_{stats}")
        else:
            for c in self.specific.keys():
                for stats in self.specific[c]:
                    columns.append(f"{c}_{stats}")
        return columns

    def _transform_all_features(
        self, X: pd.DataFrame, X_new: pd.DataFrame, rolling, abs_rolling
    ):
        for stats in self.to_compute:
            columns_to_assign = [f"{c}_{stats}" for c in X.columns]
            out = getattr(self, f"_{stats}")(X, rolling, abs_rolling)
            X_new.loc[:, columns_to_assign] = out.values

    def _transform_specific(
        self, X: pd.DataFrame, X_new: pd.DataFrame, rolling, abs_rolling
    ):
        for c in self.specific.keys():
            for stats in self.specific[c]:
                feature = f"{c}_{stats}"
                out = getattr(self, f"_{stats}")(X[c], rolling[c], abs_rolling[c])
                X_new.loc[:, feature] = out.values

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Compute features from the given life

        Parameters:
            X: The input life

        Returns:
            A new DataFrame with one row and with n columns. Let m be the number of features of the life and f the len(to_compute), then n = m x f
        """
        columns = self._compute_column_names(X)

        X_new = pd.DataFrame(index=X.index, columns=columns)
        rolling = X.rolling(self.window, self.min_points)
        abs_rolling = X.abs().rolling(self.window, self.min_points)
        if self.to_compute is not None:
            self._transform_all_features(X, X_new, rolling, abs_rolling)
        else:
            self._transform_specific(X, X_new, rolling, abs_rolling)
        return X_new


class ExpandingStatistics(TransformerStep):
    """Compute diverse number of features using an expandign window

    For each feature present in the life a number of feature will be computed for each time stamp

    The possible features are:

    - Kurtosis
    - Skewness
    - Max
    - Min
    - Std
    - Peak
    - Impulse
    - Clearance
    - RMS
    - Shape
    - Crest
    - Hurst


    Parameters:
        min_points: The minimun number of points of the expanding window, by default 2
        to_compute: List of the features to compute, by default None. Valid values are: 'kurtosis', 'skewness', 'max', 'min', 'std', 'peak', 'impulse','clearance', 'rms', 'shape', 'crest', 'hurst'
        name: Name of the step, by default None

    """

    def __init__(
        self,
        *,
        min_points: int=2,
        to_compute: List[str] = None,
        specific: Optional[Dict[str, List[str]]] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.min_points = min_points
        valid_stats = [
            "kurtosis",
            "skewness",
            "max",
            "min",
            "std",
            "peak",
            "impulse",
            "clearance",
            "rms",
            "shape",
            "crest",
            "mean",
            "deviance",
            "std_atan",
            "energy",
            "std_acosh",
            "std_asinh",
        ]
        not_default = ["energy", "deviance"]
        if to_compute is not None and specific is not None:
            raise ValueError("Only one of to_compute or specific should be used")
        self.specific = specific
        self.to_compute = to_compute
        if to_compute is None:
            if specific is None:
                self.to_compute = list(set(valid_stats) - set(not_default))
            else:
                self.specific = specific
        else:
            for f in to_compute:
                if f not in valid_stats:
                    raise ValueError(
                        f"Invalid feature to compute {f}. Valids are {valid_stats}"
                    )
            self.to_compute = to_compute

    def partial_fit(self, X, y=None):
        return self

    def fit(self, X, y=None):
        return self

    def _std_asinh(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return x.apply(np.arcsinh).expanding(self.min_points).std(numeric_only=True)

    def _std_acosh(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return x.apply(np.arccosh).expanding(self.min_points).std(numeric_only=True)

    def _energy(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return x.pow(2).expanding(self.min_points).sum(numeric_only=True)

    def _std_atan(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return x.apply(np.arctan).expanding(self.min_points).std(numeric_only=True)

    def _kurtosis(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.kurt(numeric_only=True)

    def _skewness(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.skew(numeric_only=True)

    def _max(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.max(numeric_only=True)

    def _min(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.min(numeric_only=True)

    def _std(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.std(numeric_only=True)

    def _peak(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.max(numeric_only=True) - s.min(numeric_only=True)

    def _impulse(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return self._peak(x, s, s_abs, s_abs_sqrt, s_sq) / s_abs.mean()

    def _deviance(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return (x - s.mean(numeric_only=True)) / (s.std(numeric_only=True) + 0.00000000001)

    def _clearance(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return self._peak(x, s, s_abs, s_abs_sqrt, s_sq) / s_abs_sqrt.mean().pow(2)

    # def _hurst(
    #    self,
    #    x: pd.Series,
    #    s: Expanding,
    #    s_abs: Expanding,
    #    s_abs_sqrt: Expanding,
    #    s_sq: Expanding,
    # ):
    #    return s.apply(lambda s: hurst_exponent(s, method="RS"))

    def _rms(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s_sq.mean(numeric_only=True).pow(1 / 2.0)

    def _mean(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.mean(numeric_only=True)

    def _shape(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return self._rms(x, s, s_abs, s_abs_sqrt, s_sq) / s_abs.mean(numeric_only=True)

    def _crest(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return self._peak(x, s, s_abs, s_abs_sqrt, s_sq) / self._rms(
            x, s, s_abs, s_abs_sqrt, s_sq
        )

    def _compute_column_names(self, X: pd.DataFrame):
        columns = []
        if self.to_compute is not None:
            for stats in self.to_compute:
                for c in X.columns:
                    columns.append(f"{c}_{stats}")
        else:
            for c in self.specific.keys():
                for stats in self.specific[c]:
                    columns.append(f"{c}_{stats}")
        return columns

    def _transform_all_features(
        self, X: pd.DataFrame, X_new: pd.DataFrame, expanding, s_abs, s_abs_sqrt, s_sq
    ):
        for stats in self.to_compute:
            columns_to_assign = [f"{c}_{stats}" for c in X.columns]
            out = getattr(self, f"_{stats}")(X, expanding, s_abs, s_abs_sqrt, s_sq)
            X_new.loc[:, columns_to_assign] = out.values

    def _transform_specific(
        self, X: pd.DataFrame, X_new: pd.DataFrame, expanding, s_abs, s_abs_sqrt, s_sq
    ):
        for c in self.specific.keys():
            for stats in self.specific[c]:
                feature = f"{c}_{stats}"
                out = getattr(self, f"_{stats}")(
                    X[c], expanding[c], s_abs[c], s_abs_sqrt[c], s_sq[c]
                )
                X_new.loc[:, feature] = out.values

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Compute features from the given life

        Parameters:
            X: The input life

        Returns:
            A new DataFrame with one row and with n columns. Let m be the number of features of the life and f the len(to_compute), then n = m x f,
        """
        
        columns = self._compute_column_names(X)

        X_new = pd.DataFrame(index=X.index, columns=columns)
        expanding = X.expanding(self.min_points)
        s_abs = X.abs().expanding(self.min_points)
        s_abs_sqrt = X.abs().pow(1.0 / 2).expanding(self.min_points)
        s_sq = X.pow(2).expanding(self.min_points)
        if self.to_compute is not None:
            self._transform_all_features(X, X_new, expanding, s_abs, s_abs_sqrt, s_sq)
        else:
            self._transform_specific(X, X_new, expanding, s_abs, s_abs_sqrt, s_sq)
        return X_new


class Difference(TransformerStep):
    """Compute the difference between two set of features

    Example:

        X[features1] - X[features2]

    Parameters:
        feature_set1: Feature list of the first group to substract
        feature_set2:Feature list of the second group to substract
        name: Name of the step, by default None
    """

    def __init__(
        self, *, feature_set1: List[str], feature_set2: List[str], name: Optional[str] = None
    ):
        super().__init__(name=name)
        if len(feature_set1) != len(feature_set2):
            raise ValueError(
                "Feature set 1 and feature set 2 must have the same length"
            )
        self.feature_set1 = feature_set1
        self.feature_set2 = feature_set2
        self.feature_names_computed = False

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life 

        Parameters:
            X: The input life

        Returns: 
            A DataFrame with two columns containing the result of the differences between the two sets of input features 
        """
        if not self.feature_names_computed:
            self.feature_set1 = [self.find_feature(X, c) for c in self.feature_set1]
            self.feature_set2 = [self.find_feature(X, c) for c in self.feature_set2]
            feature_names_computed = True
        new_X = X[self.feature_set1].copy()
        new_X = new_X - X[self.feature_set2].values
        return new_X


class EMD(TransformerStep):
    """Compute the empirical mode decomposition of each feature

    Parameters:
        n: Number of modes to compute
        name: Name of the step, by default None
    """

    def __init__(self, *, n: int, name: Optional[str] = "EMD"):
        super().__init__(name=name)
        self.n = n

    def transform(self, X:pd.DataFrame) -> pd.DataFrame:
        """ Apply transformation to the input life
        
        Parameters:
            X: The input life

        Returns:
            A DataFrame where the number of columns is n times the one of the input life, since each features is substituted by the n modes of its EMD

        """
        new_X = pd.DataFrame(index=X.index)
        for c in X.columns:
            try:
                imf = emd.sift.sift(X[c].values, max_imfs=self.n)
                for j in range(self.n):
                    if j < imf.shape[1]:
                        new_X[f"{c}_{j}"] = imf[:, j]
                    else:
                        new_X[f"{c}_{j}"] = np.nan
            except Exception as e:
                for j in range(self.n):
                    new_X[f"{c}_{j}"] = np.nan

        return new_X


class SlidingNonOverlappingEMD(TransformerStep):

    def __init__(
        self, *, window_size: int, max_imfs: int, keep: Optional[int] = None, **kwargs
    ):
        super().__init__(*kwargs)
        self.window_size = window_size
        self.strides = window_size
        self.max_imfs = max_imfs
        if keep is None:
            keep = self.max_imfs
        assert keep <= self.max_imfs
        self.keep = keep

    def transform(self, X: pd.DataFrame):
        def _emd(values: np.ndarray):
            out = np.zeros((values.shape[0], self.keep))
            try:
                v = emd.sift.sift(values, max_imfs=self.max_imfs)
                out[:, : self.keep] = v[:, : self.keep]
            except emd.support.EMDSiftCovergeError:
                pass

            return out

        column_list = []
        for c in X.columns:
            for i in range(self.keep):
                column_list.append(f"imf_{i}_{c}")
        out = pd.DataFrame(index=X.index, columns=column_list, dtype=np.float64)
        for c in X.columns:
            emd_computed = apply_rolling_data(
                X[c].values, _emd, self.window_size, self.strides
            )

            out.loc[:, [f"imf_{i}_{c}" for i in range(self.keep)]] = emd_computed
        return out


class EMDFilter(TransformerStep):
    """
    Filter the signals using Empirical Mode decomposition

    Parameters:
        n: Number of modes
        min_imf: Min Intrinsic Mode Function
        max_imf: Max Intrinsic Mode Function
    """

    def __init__(
        self, *, n: int, min_imf: int, max_imf: int, name: Optional[str] = "EMD"
    ):
        super().__init__(name=name)
        self.n = n
        self.min_imf = min_imf
        self.max_imf = max_imf

    def transform(self, X:pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life

        Parameters:
            X: The input life 

        Returns: 
            A DataFrame with the same shape of the input life and with the result of the EMD Filter application
        """
        new_X = pd.DataFrame(index=X.index)

        for c in X.columns:
            try:
                imf = emd.sift.sift(X[c].values, max_imfs=self.n)
                new_X[c] = np.sum(imf[:, self.min_imf : self.max_imf], axis=1)
            except Exception as e:
                new_X[c] = X[c]

        return new_X


class ChangesDetector(TransformerStep):
    """Compute how many changes there are in a categorical variable


    ['a', 'a', 'b', 'c] -> [0, 0, 1, 1]
    """

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Appply the transformation to the input life

        Parameters:
            X: The input life 
        
        Returns:
            A DataFrame with boolean values representing weather changes were applied to the input variable or not
        """
        return X != X.shift(axis=0)


class Interactions(TransformerStep):
    """Compute pairwise interactions between the features"""

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to one life 

        Parameters:
            X: The input life

        Returns:
            DataFrame containing the pairwise interaction values 
        
        """
        X_new = pd.DataFrame(index=X.index)
        for c1, c2 in itertools.combinations(X.columns, 2):
            X_new[f"{c1}_{c2}"] = X[c1] * X[c2]
        return X_new


class ROCKET(SKLearnTransformerWrapper):
    def __init__(
        self,
        *,
        n_kernels=1000,
        kernel_sizes=(7, 9, 11),
        random_state=None,
        name: Optional[str] = "ROCKET",
    ):
        transformer = pyts_ROCKET(
            n_kernels=n_kernels, kernel_sizes=kernel_sizes, random_state=random_state
        )
        super().__init__(transformer=transformer, name=name)

    def _column_names(self, X):
        a = ["Filter {i}" for i in range(self.transformer.n_kernels * 2)]
        return a


class DoG(TransformerStep):
    def __init__(
        self,
        *,
        window_size: int,
        std: List[float],
        min_points: int = 1,
        center: bool = False,
        **kwargs,
    ):
        super().__init__(*kwargs)
        self.window_size = window_size
        self.std = std
        self.center = center
        self.min_points = min_points

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X1 = X.rolling(
            window=self.window_size,
            win_type="gaussian",
            center=self.center,
            min_periods=self.min_points,
        ).mean(std=self.std[0])
        X2 = X.rolling(
            window=self.window_size,
            win_type="gaussian",
            center=self.center,
            min_periods=self.min_points,
        ).mean(std=self.std[1])
        return X1 - X2


def wrcoef(X, coef_type, coeffs, wavename, level):
    N = np.array(X).size
    a, ds = coeffs[0], list(reversed(coeffs[1:]))

    if coef_type == "a":
        return pywt.upcoef("a", a, wavename, level=level)[:N]
    elif coef_type == "d":
        return pywt.upcoef("d", ds[level - 1], wavename, level=level)[:N]
    else:
        raise ValueError("Invalid coefficient type: {}".format(coef_type))


class SlidingNonOverlappingWaveletDecomposition(TransformerStep):
    """

    # TODO TEST
    X = signal
    coeffs = pywt.wavedec(X, 'db1', level=level)
    A4 = wrcoef(X, 'a', coeffs, 'db1', level)
    D4 = wrcoef(X, 'd', coeffs, 'db1', level)
    D3 = wrcoef(X, 'd', coeffs, 'db1', 3)
    D2 = wrcoef(X, 'd', coeffs, 'db1', 2)
    D1 = wrcoef(X, 'd', coeffs, 'db1', 1)
    r = A4 + D4 + D3 + D2 + D1
    assert(np.mean(r-X) < 0.00000)

    Parameters
    ----------
    TransformerStep : [type]
        [description]
    """

    def __init__(
        self, *, window_size: int, level: int, wavelet: str, keep: List[str], **kwargs
    ):
        super().__init__(*kwargs)
        self.wavelet = wavelet
        self.level = level
        self.keep = keep
        self.window_size = window_size
        self.strides = window_size

    def transform(self, X: pd.DataFrame):
        def _wavelet(values: np.ndarray):
            coeffs = pywt.wavedec(values, self.wavelet, level=self.level)
            out = np.zeros((values.shape[0], len(self.keep)))
            for i, s in enumerate(self.keep):
                part, level = s
                out[:, i] = wrcoef(
                    values, part.lower(), coeffs, self.wavelet, int(level)
                )
            return out

        column_list = []
        for c in X.columns:
            for name in self.keep:
                column_list.append(f"wavelet_{name}_{c}")
        out = pd.DataFrame(index=X.index, columns=column_list, dtype=np.float32)
        for c in X.columns:
            wv_computed = apply_rolling_data(
                X[c].values, _wavelet, self.window_size, self.strides
            )

            out.loc[:, [f"wavelet_{name}_{c}" for name in self.keep]] = wv_computed
        return out
